# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
#
import os
import re
import shutil
import numpy as np
from math import sqrt
from qepy import qepyenv
from .pseudo import get_pseudo_path
from .tools import fortran_bool
from .lattice import red_car, car_red

class PwIn(object):
    """
    Class to generate an manipulate Quantum Espresso input files
    Can be initialized either reading from a file or starting from a new file.
    This class is not meant to be comprehensive but a lightweight version 
    capable of basic input/ouput of PW files.
    For a comprehensive class use the tools provoded by the AiiDa package: http://www.aiida.net/

    Examples of use:

    To read a local file with name "mos2.in"

        .. code-block :: python

            qe = PwIn.from_file('mos2.scf')
            print qe

    To start a file from scratch

        .. code-block :: python

            qe = PwIn()
            qe.atoms = [['N',[ 0.0, 0.0,0.0]],
                        ['B',[1./3,2./3,0.0]]]
            qe.atypes = {'B': [10.811, "B.pbe-mt_fhi.UPF"],
                         'N': [14.0067,"N.pbe-mt_fhi.UPF"]}

            qe.control['prefix'] = "'%s'"%prefix
            qe.control['verbosity'] = "'high'"
            qe.control['wf_collect'] = '.true.'
            qe.control['pseudo_dir'] = "'../pseudos/'"
            qe.system['celldm(1)'] = 4.7
            qe.system['celldm(3)'] = layer_separation/qe.system['celldm(1)']
            qe.system['ecutwfc'] = 60
            qe.system['occupations'] = "'fixed'"
            qe.system['nat'] = 2
            qe.system['ntyp'] = 2
            qe.system['ibrav'] = 4
            qe.kpoints = [6, 6, 1]
            qe.electrons['conv_thr'] = 1e-8

            print qe

    Special care should be taken with string variables e.g. "'high'"

    """
    _pw = 'pw.x'

    def __init__(self):
        """ TODO: specify the required parameters """
        #kpoints
        self.filename="pwscf.in"
        self.ktype = "automatic"
        self.kpoints = [1,1,1]
        self.shiftk = [0,0,0]
        self.klist = []
        #dictionaries
        self.control = dict(prefix="'pw'",wf_collect='.true.',verbosity="'high'")
        self.system = dict()
        self.electrons = dict(conv_thr=qepyenv.CONV_THR)
        self.ions = dict()
        self.cell = dict()
        self.atypes = dict()
        self.cell_units = 'bohr'
        self.atomic_pos_type = 'crystal'
        self.Utype = None

    @classmethod
    def from_file(cls,filename):
        """ Initialize the QE structure from a file """
        new = cls()

        with open(filename,"r") as f:
            new.filename=filename
            new.file_lines = f.readlines() #set file lines
            new.store(new.control,"control")     #read &control
            new.store(new.system,"system")      #read &system
            new.store(new.electrons,"electrons")   #read &electrons
            new.store(new.ions,"ions")        #read &ions
            new.store(new.cell,"cell")        #read &ions
            #read ATOMIC_SPECIES
            new.read_atomicspecies()
            #read ATOMIC_POSITIONS
            new.read_atoms()
            #read K_POINTS
            new.read_kpoints()
            #read CELL_PARAMETERS
            new.read_cell_parameters()
            #read HUBBARD
            new.read_hubbard()

        return new

    @classmethod
    def from_structure_dict(cls,structure,kpoints=None,ecut=None,pseudo_dir='.',conv_thr=None):
        pwi = cls()
        pwi.set_structure(structure)
        if kpoints: pwi.set_kpoints(kpoints)
        if ecut: pwi.set_ecut(ecut)
        if pseudo_dir: pwi.pseudo_dir = pseudo_dir
        if conv_thr: pwi.electrons['conv_thr'] = conv_thr
        return pwi

    @property
    def natoms(self):
        return len(self.atoms)

    @property
    def ntyp(self):
        return int(self.system['ntyp'])

    @property
    def pseudo_dir(self):
        if 'pseudo_dir' not in self.control: return None
        return self.control['pseudo_dir'].replace("'",'')

    @pseudo_dir.setter
    def pseudo_dir(self,value):
        self.control['pseudo_dir'] = "'%s'"%value.replace("'",'')

    @property
    def prefix(self): 
        return self.control['prefix'].replace("'",'')

    @prefix.setter
    def prefix(self,value):
        self.control['prefix'] = "'%s'"%value.replace("'",'')

    def set_ecut(self,ecut):
        self.system['ecutwfc'] = ecut

    def set_structure(self,structure):
        """
        Set the structure from a structure dictionary

        Example:
            .. code-block :: python

            structure = dict(ibrav=4,celldm1=4.7,celldm3=12)
            atypes = dict(Si=[28.086,"Si.pbe-mt_fhi.UPF"])
            atoms = [['N',[ 0.0, 0.0,0.5]],
                     ['B',[1./3,2./3,0.5]]]
            BN = dict(structure=structure,atypes=atypes,atoms=atoms)

            pwi = PwIn()
            pwi.set_structure(BN)
        """
        if 'lattice' in structure: self.set_lattice(**structure['lattice'])
        if 'atypes'  in structure: self.set_atypes(structure['atypes'])
        if 'atoms'   in structure: self.set_atoms(structure['atoms'])
        if 'occupations' in structure: self.set_occupations(structure['occupations'])

    def get_structure(self):
        """ Return an instance of a structure dictionary
        """
        lattice = self.get_lattice()
        return dict(lattice=lattice,atypes=self.atypes,atoms=self.atoms)

    def change_cell_parameters(self):
        """
        Convert the atomic positions to cartesian, change the lattice and convert
        the atomic positions to reduced
        """
        raise NotImplementedError('TODO')

    def set_lattice(self,ibrav=None,celldm1=None,celldm2=None,celldm3=None,
                      celldm4=None,celldm5=None,celldm6=None,cell_parameters=None):
        """Set the structure using the typical QE input variables"""
        if ibrav == 0 and cell_parameters is None:
            raise ValueError('ibrav = 0 implies that the cell_parameters variable is set')
        if cell_parameters: self.cell_parameters = cell_parameters
        if ibrav is not None: self.set_ibrav(ibrav)
        if celldm1 is not None: self.system['celldm(1)'] = celldm1
        if celldm2 is not None: self.system['celldm(2)'] = celldm2
        if celldm3 is not None: self.system['celldm(3)'] = celldm3
        if celldm4 is not None: self.system['celldm(4)'] = celldm4
        if celldm5 is not None: self.system['celldm(5)'] = celldm5
        if celldm6 is not None: self.system['celldm(6)'] = celldm6

    def get_lattice(self):
        lattice_dict = {}
        if 'ibrav' in self.system: lattice_dict['ibrav'] = self.ibrav 
        if 'celldm(1)' in self.system: lattice_dict['celldm1'] = self.system['celldm(1)']
        if 'celldm(2)' in self.system: lattice_dict['celldm2'] = self.system['celldm(2)']
        if 'celldm(3)' in self.system: lattice_dict['celldm3'] = self.system['celldm(3)']
        if 'celldm(4)' in self.system: lattice_dict['celldm4'] = self.system['celldm(4)']
        if 'celldm(5)' in self.system: lattice_dict['celldm5'] = self.system['celldm(5)']
        if 'celldm(6)' in self.system: lattice_dict['celldm6'] = self.system['celldm(6)']
        if self.ibrav == 0: lattice_dict['cell_parameters'] = self.cell_parameters
        return lattice_dict

    def set_atoms(self,atoms,coordtype="reduced"):
        """
        Set the atoms.
        Internally the atoms are always stored in reduced coordinates.
        The positions are written/read in the format from atomic_pos_type

        Arguments:
            coordtype: ("reduced"/"crystal") or cartesian

        Example:
            .. code-block :: python
   
                pwi = PwIn()
                pwi.set_atoms( [['Si',[0.125,0.125,0.125]],
                                ['Si',[-.125,-.125,-.125]]])
        """
        self.system['nat'] = len(atoms)
        #TODO: add consistency check 

        #convert the positions if needed
        if coordtype in ["reduced","crystal"]:
            self._atoms = atoms
        else:
            red_atoms = []
            for atype,apos in atoms:
                red_atoms.append( [atype,car_red([apos],self.cell_parameters)[0]] )
            self._atoms = red_atoms 

    def get_alat0(self):
        if self.system['celldm(1)'] == None:
            alat0 = np.linalg.norm(self.cell_parameters[0])
        else:
            alat0 = float(self.system['celldm(1)'])
        return alat0


    def get_atoms(self, units=None):
        from .units     import ang2au,au2ang
        from .lattice   import red_car,car_red

        atoms_arr= np.array([atom[1] for atom in self.atoms])   #atom[0] is the atomic symbol; atom[1] is the 3 coord list
        units = units if units is not None else self.atomic_pos_type   #self.atomic_pos_type = crystal in my case

        if units == self.atomic_pos_type:
            return self.atoms

        scale_in =1.0
        if self.atomic_pos_type == "angstrom":
            scale_in = ang2au
        elif self.atomic_pos_type == "alat":
            if self.system['celldm(1)'] == None:
                scale_in=np.linalg.norm(self.cell_parameters[0])
            else:
                scale_in = float(self.system['celldm(1)'])
        elif self.atomic_pos_type == "crystal":
            atoms_arr = red_car(atoms_arr, np.array(self.cell_parameters))

        atoms_arr*=scale_in # transform in bohr

        scale_out=1.0
        if units == "alat":
            if self.system['celldm(1)'] == None:
                scale_out=1.0/np.linalg.norm(self.cell_parameters[0])
            else:
                scale_out=1.0/float(self.system['celldm(1)'])
        elif units == "angstrom":
            scale_out=au2ang
        elif units == "crystal":
            atoms_arr = car_red(atoms, np.array(self.cell_parameters))

        atoms_arr*=scale_out # transform in bohr
        atoms_string=""
        for atom,arr in zip(self.atoms,atoms_arr):
            atoms_string+="%3s %14.10lf %14.10lf %14.10lf \n" % (atom[0], arr[0], arr[1], arr[2])
        return atoms_string

    def set_hubbard(self,Uvalues,Utype="ortho-atomic"):
        """
        Set the Hubbard corrections: [ ion-orbital, U(eV) ]

        Example:
            .. code-block :: python

                pwi = PwIn()
                pwi.set_hubbard( [['Ce-4f', 5.0 ],
                                  ['O-2p, 1e-4 ]])
        """
        self.Uvalues = Uvalues
        self.Utype   = Utype

    def set_atypes(self,atypes):
        """"
        Set the atom types.

        Example:
            .. code-block :: python
                pwi = PwIn()
                pwi.set_atypes({'Si': [28.086,"Si.pbe-mt_fhi.UPF"]})
        """
        self.system['ntyp'] = len(atypes)
        #TODO: add consistency check
        self.atypes = atypes

    def set_occupations(self,occupations):
        self.system['occupations'] = "'%s'" % occupations['occupations'] 
        
        if 'smearing' in occupations: self.system['smearing'] = "'%s'" % occupations['smearing']
        if 'degauss'  in occupations: self.system['degauss']  = occupations['degauss'] 
        if 'nbnd'     in occupations: self.system['nbnd']     = occupations['nbnd']

    def update_structure_from_xml(self,pwxml):
        """
        Update input from an PW xml file. Useful for relaxation.
        """
        structure_dict = pwxml.get_structure_dict()
        self.set_structure(structure_dict)

    def set_nscf(self,nbnd,nscf_kpoints=None,conv_thr=1e-8,
                 diago_full_acc=True,force_symmorphic=True):
        """
        set the calculation to be nscf
        """
        self.control['calculation'] = "'nscf'"
        self.electrons['conv_thr'] = conv_thr
        self.system['nbnd'] = nbnd
        self.electrons['diago_full_acc'] = fortran_bool(diago_full_acc)
        self.system['force_symmorphic'] = fortran_bool(force_symmorphic)
        if nscf_kpoints: self.set_kpoints(nscf_kpoints) 
        return self

    def set_bands(self,nbnd,path_kpoints=None,conv_thr=1e-8,
                 diago_full_acc=True,force_symmorphic=True):
        """
        set the calculation to be nscf
        """
        self.control['calculation'] = "'bands'"
        self.electrons['conv_thr'] = conv_thr
        self.system['nbnd'] = nbnd
        self.electrons['diago_full_acc'] = fortran_bool(diago_full_acc)
        self.system['force_symmorphic'] = fortran_bool(force_symmorphic)
        self.ktype = 'crystal'
        if path_kpoints: self.set_path(path_kpoints) 
        return self

    def set_relax(self,cell_dofree=None):
        """
        set the calculation to be relax
        """
        self.control['calculation'] = "'relax'"
        self.ions['ion_dynamics']  = "'bfgs'"
        if cell_dofree:
            self.control['calculation'] = "'vc-relax'"
            self.cell['cell_dynamics']  = "'bfgs'"
            self.cell['cell_dofree'] = "'%s'"%cell_dofree
        return self

    def set_spinorbit(self):
        self.system['lspinorb'] = '.true.'
        self.system['noncolin'] = '.true.'

    def set_spinpolarized(self):
        self.system['lspinorb'] = '.false.'
        self.system['noncolin'] = '.false.'
        self.system['nspin']    = 2

    def set_magnetization(self,starting_magnetization):
        """
        Set the starting_magnetization for a spin calculation.

        Args:
            starting_magnetization: a list with the starting magnetizations for each atomic type
        """
        if starting_magnetization is None: return
        if len(starting_magnetization) != self.ntyp:
            raise ValueError('Invalid size for starting_magnetization')
        for atom_type,magnetization in enumerate(starting_magnetization):
            self.system['starting_magnetization(%d)' % (atom_type+1)] = magnetization

    def get_pseudos(self,destpath='.',pseudo_paths=[],verbose=0):
        """
        Check if the pseudopotential files can be found in the specified path
        and copy them to destpath
        """
        import qepy.data.pseudos as qe_pseudos
        pseudo_paths.append(os.path.dirname(qe_pseudos.__file__))

        #use pseudo_dir from control
        if self.pseudo_dir:
            if os.path.isdir(self.pseudo_dir): 
                pseudo_paths.append(self.pseudo_dir)

        ppstring = '\n'.join(pseudo_paths)
        if verbose: print('List of pseudo_paths:\n'+ppstring)

        #find all the pseudopotentials
        for atype,(mass,pseudo_filename) in self.atypes.items():
            pseudo_filename = get_pseudo_path( pseudo_filename, pseudo_paths )
            if pseudo_filename is None:
                err_msg = 'Pseudopotential %s not found in any of these paths:\n'%pseudo_filename
                raise ValueError(err_msg+ppstring)

            if verbose: print('cp %s %s'%(pseudo_filename,destpath))
            shutil.copy(pseudo_filename,destpath)

    def set_kpoints(self,kpoints,shiftk=[0,0,0]):
        """Add some logic to set the kpoints mesh"""
        #sanity check
        if len(kpoints) != 3: raise ValueError('Wrong kpoints dimensions')
        self.kpoints = kpoints
        self.shiftk = shiftk

    def copy(self):
        """Return a copy of this instance"""
        import copy
        return copy.deepcopy(self)

    def read_atomicspecies(self):
        lines = iter(self.file_lines)
        #find ATOMIC_SPECIES keyword in file and read next line
        for line in lines:
            if "ATOMIC_SPECIES" in line:
                for i in range(self.ntyp):
                    atype, mass, psp = next(lines).split()
                    self.atypes[atype] = [mass,psp]

    def get_symmetry_spglib(self):
        """
        get the symmetry group of this system using spglib
        """
        import spglib

        lat, positions, atypes = self.get_cell()
        lat = np.array(lat)

        at = np.unique(atypes)
        an = dict(list(zip(at,list(range(len(at))))))
        atypes = [an[a] for a in atypes]

        cell = (lat,positions,atypes)

        spacegroup = spglib.get_spacegroup(cell,symprec=1e-5)
        return spacegroup

    def get_masses(self):
        """ Get an array with the masses of all the atoms
        """
        masses = []
        for atom in self.atoms:
            atype = self.atypes[atom[0]]
            mass = float(atype[0])
            masses.append(mass)
        return masses

    def set_path(self,path):
        self.klist = path.get_klist()

    def get_cell(self):
        """ Get the lattice parameters, positions of the atoms and chemical symbols
        """
        cell = self.cell_parameters
        sym = [atom[0] for atom in self.atoms]
        pos = [atom[1] for atom in self.atoms]
        if self.atomic_pos_type == 'bohr':
            pos = car_red(pos,cell)
        return cell, pos, sym

    def set_atoms_string(self,string):
        """
        set the atomic positions using string of the form
        Si 0.0 0.0 0.0
        Si 0.5 0.5 0.5
        """
        atoms_str = [line.strip().split() for line in string.strip().split('\n')]
        atoms = []
        for atype,x,y,z in atoms_str:
            atoms.append([atype,list(map(float,[x,y,z]))])
        self.atoms = atoms

    def set_atoms_ase(self,atoms):
        """ set the atomic positions using a Atoms datastructure from ase
        """
        # we will write down the cell parameters explicitly
        self.ibrav = 0
        self.cell_parameters = atoms.get_cell()
        self.atoms = list(zip(atoms.get_chemical_symbols(),atoms.get_scaled_positions()))

    def displace(self,cart_mode,displacement,masses=None):
        """ Displace the atoms acoording to a phonon mode """
        import copy
        #normalize with masses
        if masses is None: masses = self.get_masses()
        if len(masses) != self.natoms: raise ValueError('Wrong dimensions of the masses list')

        #apply displacement
        atomic_car_pos_0 = self.atomic_car_pos
        atoms_car_pos_disp = []
        for i,(atype,apos) in enumerate(self.atoms):
            delta = cart_mode[i].real*displacement/sqrt(masses[i])
            atom = [atype, atomic_car_pos_0[i]+delta]
            atoms_car_pos_disp.append( atom )

        #store displaced structures
        self.set_atoms( atoms_car_pos_disp, coordtype="cartesian" )
        
    @property    
    def atomic_red_pos(self):
        pos = []
        for i in range(self.natoms):
            pos.append(self.atoms[i][1])
        return pos

    @property
    def atomic_car_pos(self):
        pos = []
        for i in range(self.natoms):
            red_pos = self.atoms[i][1]
            pos.append(red_car([red_pos],self.cell_parameters)[0])
        return pos

    @property
    def atoms(self):
        return self._atoms

    def get_displaced(self,cart_mode,displacement,masses=None):
        """ Create a copy and displace along the phonon mode """
        copy = self.copy()
        copy.displace(cart_mode,displacement,masses=masses)
        return copy

    def read_atoms(self):
        lines = iter(self.file_lines)
        #find READ_ATOMS keyword in file and read next lines
        for line in lines:
            if "ATOMIC_POSITIONS" in line:
                atomic_pos_type = line
                self.atomic_pos_type = re.findall('([A-Za-z]+)',line)[-1]
                atoms = []
                for i in range(int(self.system["nat"])):
                    atype, x,y,z = next(lines).split()
                    atoms.append([atype,[float(i) for i in (x,y,z)]])
        self._atoms = atoms
        self.atomic_pos_type = atomic_pos_type.replace('{','').replace('}','').strip().split()[1]

    def read_hubbard(self):
        lines = iter(self.file_lines)
        #find HUBBARD keyword in file and read next lines
        for line in lines:
            if "HUBBARD" in line:
                self.Utype = line.replace('{','').replace('}','').strip().split()[1]
                self.Uvalues = []
            if self.Utype is not None and line.startswith('U'):
                orbital = line.strip().split()[1]
                UeV = float(line.strip().split()[2])
                self.Uvalues.append([ orbital, UeV  ])

    @property
    def alat(self):
        if self.ibrav == 0:
            return np.linalg.norm(self.cell_parameters[0])
        else:
            return self.system['celldm(1)']

    @property
    def cell_parameters(self):
        if self.ibrav == 0:
            return self._cell_parameters
        elif self.ibrav == 1:
            a = float(self.system['celldm(1)'])
            cell_parameters = [[  a,   0,   0],
                               [  0,   a,   0],
                               [  0,   0,   a]]
        elif self.ibrav == 2:
            a = float(self.system['celldm(1)'])
            cell_parameters = [[ -a/2,   0, a/2],
                               [    0, a/2, a/2],
                               [ -a/2, a/2,   0]]
        elif self.ibrav == 3:
            a = float(self.system['celldm(1)'])
            cell_parameters = [[ a/2,  a/2,  a/2],
                               [-a/2,  a/2,  a/2],
                               [-a/2, -a/2,  a/2]]
        elif self.ibrav == 4:
            a = float(self.system['celldm(1)'])
            c = float(self.system['celldm(3)'])
            cell_parameters = [[   a,          0,  0],
                               [-a/2,sqrt(3)/2*a,  0],
                               [   0,          0,c*a]]
        elif self.ibrav == 6:
            a = float(self.system['celldm(1)'])
            c = float(self.system['celldm(3)'])
            cell_parameters = [[  a,   0,   0],
                               [  0,   a,   0],
                               [  0,   0, c*a]]
        elif self.ibrav == -5:
            a = float(self.system['celldm(1)'])
            c = float(self.system['celldm(4)'])
            ap = a/sqrt(3.0)
            ty = sqrt((1.0-c)/6.0)
            tz = sqrt((1.0+2.0*c)/3.0)
            u = tz - 2.0*sqrt(2.0)*ty
            v = tz + sqrt(2.0)*ty
            cell_parameters = [[  ap*u, ap*v, ap*v],
                               [  ap*v, ap*u, ap*v],
                               [  ap*v, ap*v, ap*u]]
            

        else:
            raise NotImplementedError('ibrav = %d not implemented'%self.ibrav)
        return cell_parameters

    @cell_parameters.setter
    def cell_parameters(self,value):
        self._cell_parameters = value

    @property
    def ibrav(self):
        return int(self.system['ibrav'])

    @ibrav.setter
    def ibrav(self,value):
       self.set_ibrav(value)

    def set_ibrav(self,value):
        if not hasattr(self,'_cell_parameters') and value == 0:
            raise ValueError('Must set cell_parameters before setting ibrav to 0')
        if value == 0:
            for i in range(1,7):
                self.remove_key(self.system,'celldm(%d)'%i)
        self.system['ibrav'] = value

    def read_cell_parameters(self):
        """read the cell parameters from the input file """
        def rmchar(string,symbols): return ''.join([c for c in string if c not in symbols])

        if self.ibrav == 0:
            if 'celldm(1)' in list(self.system.keys()):
                a = float(self.system['celldm(1)'])
            else:
                a = 1
            lines = iter(self.file_lines)
            for line in lines:
                if "CELL_PARAMETERS" in line:
                    units = rmchar(line.strip(),'{}()').split()
                    cell_parameters = [[],[],[]]
                    if len(units) > 1:
                        self.cell_units = units[1]
                    else:
                        self.cell_units = 'bohr'
                    for i in range(3):
                        cell_parameters[i] = [ float(x)*a for x in next(lines).split() ]
            if self.cell_units == 'angstrom' or self.cell_units == 'bohr':
                if 'celldm(1)' in self.system: del self.system['celldm(1)']
            if 'celldm(1)' not in list(self.system.keys()):
                a = np.linalg.norm(cell_parameters[0])
            self._cell_parameters = cell_parameters

    def read_kpoints(self):
        lines = iter(self.file_lines)
        #find K_POINTS keyword in file and read next line
        for line in lines:
            if "K_POINTS" in line:
                #chack if the type is automatic
                if "automatic" in line:
                    self.ktype = "automatic"
                    vals = list(map(float, next(lines).split()))
                    self.kpoints, self.shiftk = vals[0:3], vals[3:6]
                #otherwise read a list
                else:
                    #read number of kpoints
                    nkpoints = int(next(lines).split()[0])
                    self.klist = []
                    try: self.ktype = line.replace('{','').replace('}','').strip().split()[1]
                    except IndexError: self.ktype = ""
                    try:
                        lines_list = list(lines)
                        for n in range(nkpoints):
                            vals   = lines_list[n].split()[:4]
                            self.klist.append( list(map(float,vals)) )    
                    except IndexError:
                        print("wrong k-points list format")
                        exit()
                    self.klist = [ [a,b,c,int(d)] for a,b,c,d in self.klist ]

    def slicefile(self, keyword):
        file_slice_regexp = rf'&{keyword}(?:.?)+\n((?:.+\n)+?)(?:\s+)?[\/&]'
        lines = re.findall(file_slice_regexp,"".join(self.file_lines),re.MULTILINE | re.IGNORECASE)
        return lines

    def store(self,group,name):
        """
        Save the variables specified in each of the groups on the structure
        """
        group_regexp = r'([a-zA-Z_0-9_\(\)]+)(?:\s+)?=(?:\s+)?([a-zA-Z\'"0-9_.+-]+)' 
        for file_slice in self.slicefile(name):
            for keyword, value in re.findall(group_regexp,file_slice):
                group[keyword.strip()]=value.strip()

    def stringify_group(self, keyword, group):
        if group != {}:
            string='&%s\n' % keyword
            for keyword in sorted(group): # Py2/3 discrepancy in keyword order
                string += "%20s = %s\n" % (keyword, group[keyword])
            string += "/"    ### /%&end does not work for all qe versions
            return string
        else:
            return ''

    def remove_key(self,group,key):
        """ if a certain key exists in the group, remove it """
        if key in list(group.keys()):
            del group[key]

    def write(self,filename):
        """write the file to disk """
        with open(filename,'w') as f:
           f.write(str(self))

    def get_string(self):
        """
        Output the file in the form of a string
        """
        lines = []; app = lines.append
        app( self.stringify_group("control",self.control) ) #print control
        app( self.stringify_group("system",self.system) ) #print system
        app( self.stringify_group("electrons",self.electrons) ) #print electrons
        app( self.stringify_group("ions",self.ions) ) #print ions
        app( self.stringify_group("cell",self.cell) ) #print ions

        #print atomic species
        app( "ATOMIC_SPECIES" )
        for atype in self.atypes:
            app( " %3s %8s %20s" % (atype, self.atypes[atype][0], self.atypes[atype][1]) )

        #print atomic positions
        app( "ATOMIC_POSITIONS { %s }"%self.atomic_pos_type )
        for atom in self.atoms:
            app( "%3s %14.10lf %14.10lf %14.10lf" % (atom[0], atom[1][0], atom[1][1], atom[1][2]) )

        #print kpoints
        if self.ktype == "automatic":
            app( "K_POINTS { %s }" % self.ktype )
            app( ("%3d"*6)%(tuple(self.kpoints) + tuple(self.shiftk)) )
        else:
            fmt_string = ""
            for element in self.klist[0]: 
                if isinstance(element,int): fmt_string+="%3d"
                else: fmt_string+="%12.8lf"
            app( "K_POINTS { %s }" % self.ktype )
            app( "%d" % len(self.klist) )
            for kpt in self.klist: app( fmt_string%tuple(kpt) )

        #print cell parameters
        if self.ibrav == 0:
            app( "CELL_PARAMETERS %s"%self.cell_units )
            app( ("%14.10lf "*3)%tuple(self.cell_parameters[0]) )
            app( ("%14.10lf "*3)%tuple(self.cell_parameters[1]) )
            app( ("%14.10lf "*3)%tuple(self.cell_parameters[2]) )

        #print hubbard
        if self.Utype is not None:
            app( "HUBBARD { %s }" % self.Utype )
            for U in self.Uvalues:
                app( "U %s %lf" % (U[0], U[1]) )

        return "\n".join(lines)


    def __str__(self):
        return self.get_string()



import numpy as np
import matplotlib.pyplot as plt

"""
Script to draw band plots.

It is called by generate_bands.

Arguments are:
    data, plt_type, show
    
    - data: [prefix,nkpoints,nbands,kstps,gaps,eigen,(shifted_eigen)] 
    - plt_type:
        -- 'bands': standard band plot with "Fulvio"-style format and layout 
    - out_name: string to be attached to output file name
    - erange: energy window centered at Fermi level (single value in eV, None means all energies)
    - show: determine if plt.show() is called at the end
    - print_data: if True, print plotted data in text format
"""

def plot_driver(data,plt_type,out_name=None,erange=None,show=True,print_data=True):
    """
    This function is like the __init__ function of a class.
    
    This script may be transformed into a class when new plot types are added
    """
    # Scissored energies are present
    if len(data)==9: prefix,nkpoints,nbands,kstps,KPTs_labels,points,gaps,eigs_noscissor,eigen = data
    # Scissored energies are not present
    elif len(data)==8: prefix,nkpoints,nbands,kstps,KPTs_labels,points,gaps,eigen = data
    # Data missing
    else: raise IndexError('Incorrect number of elements in plot data.')

    if eigen.shape[1] != nbands:
        if eigen.shape[1] == 2*nbands: eigen = np.array( [ eigen[:,:nbands],  eigen[:,nbands:] ]  ) 
        else: raise ValueError("Mismatch between nbands and eigenvalue array shape")

    if plt_type=='bands': 
        data_to_plot = prefix,nkpoints,nbands,kstps,KPTs_labels,points,gaps,eigen
        electron_dispersion_plot(data_to_plot,out_name,erange,show)
    if print_data: print_out_files(kstps,eigen,prefix,out_name)

def print_out_files(kstps,eigen,prefix,out_nm):
    """
    Print data as *.dat file
    """

    # Output name
    if out_nm is not None: out_file_dat = "%s_%s.dat"%(prefix,out_nm)
    else:                  out_file_dat = "%s_bands.dat"%prefix

    l_magnetic = (len(eigen.shape)==3)

    # Create array to print
    if not l_magnetic:
        Nk,Nb = eigen.shape
        to_prnt = np.zeros((Nk,Nb+1))
        to_prnt[:,0] =kstps
        to_prnt[:,1:]=eigen

        # Save array
        np.savetxt(out_file_dat,to_prnt,fmt='%.6f')

    if l_magnetic:
        Nspin,Nk,Nb = eigen.shape
        to_prnt_up = np.zeros((Nk,Nb+1))
        to_prnt_up[:,0] =kstps
        to_prnt_up[:,1:]=eigen[0]
        to_prnt_dn = np.zeros((Nk,Nb+1))
        to_prnt_dn[:,0] =kstps
        to_prnt_dn[:,1:]=eigen[1]

        # Save arrays
        np.savetxt('SPIN1_'+out_file_dat,to_prnt_up,fmt='%.6f')
        np.savetxt('SPIN2_'+out_file_dat,to_prnt_dn,fmt='%.6f')

def electron_dispersion_plot(data,out_nm,erange,show):
    """
    Standard band plot with "Fulvio"-style format and layout
    """
    from math import floor,ceil
    from matplotlib import rc,rcParams

    # Fonts in standard TeX style
    rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
    rc('text', usetex=True)

    # Data
    prefix,nkpoints,nbands,kstps,KPTs_labels,KPTs,gaps,eigen = data

    # Output name
    if out_nm is not None: out_file_pdf = "%s_%s.pdf"%(prefix,out_nm)
    else:                  out_file_pdf = "%s_bands.pdf"%prefix

    l_magnetic = (len(eigen.shape)==3)

    # Initial preparations
    if gaps[0]<1.e-6: yref=0.
    else:             yref=min(gaps)
    if not l_magnetic: Nk,Nb = eigen.shape
    if l_magnetic:  Ns,Nk,Nb = eigen.shape
    ylims = np.array([yref-erange,yref+erange])/2.
    xlims = [kstps[0],kstps[-1]]
    for il in range(len(KPTs_labels)):
        if KPTs_labels[il]=='G': KPTs_labels[il]=r'$\Gamma$'
    ## linewidths
    band_linewidth  = 1.5
    frame_linewidth = 3./4.*band_linewidth
    faint_linewidth = band_linewidth/3.
    rcParams['axes.linewidth'] = frame_linewidth
    ## yticks
    l_yticks = True
    N_ticks = floor(ylims[1])-ceil(ylims[0])
    if erange is None or erange>=20.: l_yticks=False
    if erange<20. and erange>=10.:
        yticks = [ ceil(ylims[0]) + tick for tick in range(N_ticks) ]
    if erange<10. and erange>=2.:
        yticks = [0.]
        tick = yticks[0]
        while tick<ylims[1]:
            tick +=0.5
            yticks.append(tick)
        tick = yticks[0]
        while tick>ylims[0]:
            tick -=0.5
            yticks.insert(0,tick)
    if erange<2.:
        yticks = [0.]
        tick = yticks[0]
        while tick<ylims[1]:
            tick +=0.1
            yticks.append(tick)
        tick = yticks[0]
        while tick>ylims[0]:
            tick -=0.1
            yticks.insert(0,tick)
    if l_yticks: yticklabels = [str(ytick) for ytick in yticks]
    ## xticks
    xticks = [0.]+KPTs
    xticklabels = KPTs_labels

    # Start plot
    fig, ax = plt.subplots(figsize=(8,6))
    ax.set_ylim(ylims)
    ax.set_ylabel(r'Energy (eV)',size=20)
    if l_yticks:
        ax.set_yticks(yticks[1:])
        ax.set_yticklabels(yticklabels[1:],size=20)
    ax.set_xlim(xlims)
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels,size=20)
    ax.tick_params(direction='in',width=frame_linewidth,length=6,left=True,right=True)
    for point in KPTs[:-1]: ax.axvline(point,color='black',linewidth=frame_linewidth)
    if yref!=0.: ax.axhline(min(gaps),color='gray',linestyle='--',linewidth=faint_linewidth)
    ax.axhline(0,color='black',linewidth=faint_linewidth)

    # Draw plot
    if not l_magnetic:
        for ib in range(Nb): ax.plot(kstps,eigen[:,ib],ls='-',lw=band_linewidth,c='red')
    if l_magnetic:
        clrs = ['red','blue']
        for i_s in range(Ns):
            for ib in range(Nb): ax.plot(kstps,eigen[i_s][:,ib],ls='-',lw=band_linewidth,c=clrs[i_s])

    plt.savefig(out_file_pdf)
    if show: plt.show()



#!/usr/bin/env python3
#
# Authors: MN
import argparse
import numpy as np
import os

from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.wfdb import YamboWFDB


def run_plot_exc_wf_real_space(args):
    parser = argparse.ArgumentParser(description="Generate real-space " +
                                     "exciton wavefunction in Gaussian .cube file.")

    parser.add_argument("--path", type=str, default=".",
                        help="Calculation directory [PATH/SAVE] (default: Current working directory '.')")
    parser.add_argument("-J","--jobdir", type=str, default="SAVE", metavar="DIR",
                        help="BSE JOB directory [PATH/DIR] (default: SAVE)")
    parser.add_argument("--iqpt", type=int, default=1,help="Q-point index (default: 1)")
    parser.add_argument("--iexe", type=int, required=True,help="Exciton index to plot. (counts from 1)")
    # --iexe and --iqpt are not python indexing. i,e 1st item starts from 1.
    parser.add_argument("--supercell", nargs=3, type=int, default=[1, 1, 1],
                        help="Supercell dimensions (default: 1 1 1)")
    parser.add_argument("--wfc_cutoff", type=float, default=-1.0,
                        help="Wavefunction cutoff in Ry (default: −1 → full cutoff)")
    parser.add_argument("--degen_tol", type=float, default=0.01,
                        help="Degeneracy threshold in eV (default: 0.01)")
    # Mutually-exclusive: fix hole OR electron position
    grp = parser.add_mutually_exclusive_group(required=True)
    grp.add_argument("--hole", nargs=3, type=float,
                     help="Fix hole position in reduced units.")
    grp.add_argument("--electron", nargs=3, type=float,
                     help="Fix electron position in reduced units.")
    #
    parser.add_argument("--phase", action="store_true",
                        help="Enable phase plot (default: False). "
                        "Works only for non-degenerate excitons "
                        "and nspin=nspinor=1.")
    #
    parser.add_argument("--block_size", type=int, default=256,
                        help="Block size for computation (default: 256). "
                        "Lower only for very tight memory.")
    #
    parser.add_argument("--neig_load", type=int, default=100,
                        help="Number of extra eigevectors to load in memory (default: 100). "
                        "Only meaningful for TDA (non TDA, it is ignored and all are read). "
                        "Default 100 is fine and in most cases can be reduced, but if you are in deep "
                        "continum, we might have more accidental degenerates, in that case "
                        "we need to load more eigenvectors.")
    #
    args = parser.parse_args(args)
    #
    calc_path = args.path
    BSE_dir = args.jobdir
    iqpt = args.iqpt
    #
    if args.hole:
        fix_particle = "h"
        fixed_position = args.hole
    else:
        fix_particle = "e"
        fixed_position = args.electron
    #
    nsfile = os.path.join(calc_path, "SAVE", "ns.db1")
    #
    lattice = YamboLatticeDB.from_db_file(nsfile)
    #
    filename = f"ndb.BS_diago_Q{iqpt}"
    # NM : In the case of TDA, we donot want to load all the eigenvectors. 
    excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
                                        folder=BSE_dir,
                                        neigs=args.iexe + args.neig_load)
    #
    wfdb = YamboWFDB(path=calc_path,latdb=lattice,
                     bands_range=[np.min(excdb.table[:, 1]) - 1,np.max(excdb.table[:, 2])])
    #
    excdb.real_wf_to_cube(iexe=args.iexe-1,wfdb=wfdb,fixed_position=fixed_position,
                          supercell=args.supercell,degen_tol=args.degen_tol,
                          wfcCutoffRy=args.wfc_cutoff,fix_particle=fix_particle,
                          phase=args.phase, block_size=args.block_size)






# Copyright (C) 2015 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy

'''
Plot the excitonic weights on the brillouin zone using ypp
Plot the absorption spectrum with the excitonic energies
'''
from __future__ import print_function

from builtins import str
from builtins import range
import matplotlib
matplotlib.use('Agg') # prevents crashes if no X server present
from yambopy import *
from qepy import *
import json
import matplotlib.pyplot as plt
import numpy as np
from math import sqrt, ceil
import argparse

parser = argparse.ArgumentParser(description='Plot excitonic weights in BZ or excitonic eigenenergies on optical spectrum using ypp.')
parser.add_argument('folder', help='Folder with SAVE and run.')
parser.add_argument('jobname', help='Job from which excitons are of interest.')
parser.add_argument('-ie','--intexc' , help='Minimum intensity for excitons to be considered bright', default=0.1)
parser.add_argument('-de','--degenexc', help='Energy threshold under which different peaks are merged (eV)', default=0.01)
parser.add_argument('-me','--maxexc', help='Energy threshold after which excitons are not read anymore (eV)', default=4.0)
parser.add_argument('-c','--cut', help='Set the limits of the BZ plot.', default=0.2)
parser.add_argument('-ny','--noypp', help='Skips the ypp calls and the .json file building.', action='store_false')
parser.add_argument('-nw','--noweight', help='Skips the plot of the weights in the BZ.', action='store_false')
parser.add_argument('-ns','--nospectrum', help='Skips the plot of the spectrum with the excitons eigenenergies.', action='store_false')
args = parser.parse_args()

jobname = args.jobname
path    = args.folder 
cut     = args.cut
exc_int   = args.intexc
exc_degen = args.degenexc
exc_max_E = args.maxexc

def get_var(dictionary,variables):
    """
    To have compatibility with different version of yambo
    We provide a list of different possible tags
    """
    for var in variables:
        if var in dictionary:
            return dictionary[var]
    raise ValueError( 'Could not find the variables %s in the output file'%str(variables) )

if args.noypp:
    print('Preparing JSON file. Calling ypp ...')
    ### Creating the 'absorptionspectra.json' file
    y = YamboOut(folder=path,save_folder=path)
    # Args : name of job, SAVE folder path, folder where job was run path
    a = YamboBSEAbsorptionSpectra(jobname,path=path)
    # Get excitons values
    a.get_excitons(min_intensity=exc_int,max_energy=exc_max_E,Degen_Step=exc_degen)
    # Get excitonic WFs
    a.get_wavefunctions(Degen_Step=exc_degen,repx=list(range(-1,2)),repy=list(range(-1,2)),repz=list(range(1)))
    # Write .json file
    a.write_json(filename=path+'_'+jobname)
else:
    print('YPP call disabled.')

### Loading data from .json file
f = open(path+'_'+jobname+'.json')
data = json.load(f)
f.close()
print('JSON file loaded.')

### Plotting the absorption spectra
print('Absorption spectra ...')
# BSE spectra
plt.plot(get_var(data,['E/ev','E/ev[1]']), get_var(data,['EPS-Im[2]' ]),label='BSE',lw=4)
# IP spectra
plt.plot(get_var(data,['E/ev','E/ev[1]']), get_var(data,['EPSo-Im[4]']),label='IP',lw=4)
# Axes : lines for exciton energies, labels
nexcitons = len(data['excitons'])
print('Number of excitons: ', nexcitons) 
for n,exciton in enumerate(data['excitons']):
    plt.axvline(exciton['energy'])
plt.xlabel('$\omega$ (eV)')
plt.ylabel('Intensity (arb. units)')
plt.gca().yaxis.set_major_locator(plt.NullLocator())
plt.legend()
#plt.draw()
#plt.show()
plt.savefig(path+'_'+jobname+'_abs.png', bbox_inches='tight')
print(path+'_'+jobname+'_abs.png')

### Lattice : necessary to determine bounds of exciton weights plot
print('Reciprocal lattice ...')
lat = np.array(data['lattice'])
rlat = rec_lat(lat)
x,y,z = np.array(rlat )
xmin,ymin,_ = -(x+y)*cut
xmax,ymax,_ = +(x+y)*cut


### Plot excitons
print('Excitons weights ...')
nx = int(ceil(sqrt(nexcitons)))
ny = int(ceil(nexcitons*1.0/nx))
print("cols:",nx)
print("rows:",ny)
cmap = plt.get_cmap("gist_heat_r")

fig = plt.figure(figsize=(nx*4,ny*4))
sorted_excitons = sorted(data['excitons'],key=lambda x: x['energy'])

for n,exciton in enumerate(sorted_excitons):
    #get data
    w   = np.array(exciton['weights'])
    qpt = np.array(exciton['qpts'])

    #plot
    ax = plt.subplot(ny,nx,n+1)
    ax.scatter(qpt[:,0], qpt[:,1], s=20, c=w, marker='H', cmap=cmap, lw=0, label="e: %lf"%exciton['energy'])

    # axis
    plt.xlim([-cut,cut])
    plt.ylim([-cut,cut])
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    ax.set_aspect('equal')

#plt.draw()
#plt.show()
plt.savefig(path+'_'+jobname+'_ex.png', bbox_inches='tight')
print('path'+'_'+jobname+'_ex.png')
print('Done.')



#
# Authors: MN
#
from yambopy.bse.exciton_irreps import compute_exc_rep
import argparse
#
def run_exc_irrep(args):
    """
    Compute the irreducible representation (irrep) of excitonic states.

    This is a command-line interface helper that parses input arguments
    and invokes :func:`compute_exc_rep` to analyze exciton symmetry
    properties from a BSE (Bethe–Salpeter Equation) calculation.

    Parameters
    ----------
    args : list[str]
        Command-line arguments list (e.g. ``sys.argv[1:]``). The supported options are:

        * ``--path`` : str, optional
          Path to the main calculation directory (default: ``"."``).

        * ``-J`` / ``--jobdir`` : str, optional
          Subdirectory containing the BSE results (default: ``"SAVE"``).

        * ``--iqpt`` : int, optional
          Index of the transferred momentum (q-point) to analyze (default: ``1``).

        * ``--nstates`` : int, optional
          Number of exciton states to process (default: ``1``).

        * ``--degen_tol`` : float, optional
          Energy tolerance used to detect degenerate excitons, in eV (default: ``1e-2``).

        * ``--sym_tol`` : float, optional
          Numerical tolerance for evaluating symmetry operations (default: ``1e-2``).

        * ``--use_save_symm`` : bool, optional
          If True, use save symmetries instead of spglib ones. For non-collinear magnets, it is automatically
          set to True,

    Notes
    -----
    This function only handles argument parsing and dispatching.
    """
    parser = argparse.ArgumentParser(description="Compute exciton representation of excitonic states.")
    #
    parser.add_argument("--path", type=str, default=".", help="Path to the calculation directory. Default current directory.")
    parser.add_argument("-J", "--jobdir", type=str, default="SAVE", metavar="DIR", help="BSE job directory. Default SAVE")
    parser.add_argument("--iqpt", type=int, default=1, help="Q-point index. Default 1")
    parser.add_argument("--nstates", type=int, default=1, help="Number of exciton states. Default 1")
    parser.add_argument("--degen_tol", type=float, default=1e-2, help="Tolerance for degeneracy. Default 0.01 eV")
    parser.add_argument("--sym_tol", type=float, default=1e-2, help="Tolerance for Symmetry operations. Default 0.01")
    parser.add_argument("--use_save_symm", action="store_true", help="use save symmetries instead of spglib ones. "
                        "For non-collinear magnets, it is automatically set to True")
    #
    args = parser.parse_args(args)
    compute_exc_rep(path=args.path,bse_dir=args.jobdir, iqpt=args.iqpt,
                    nstates=args.nstates, degen_tol=args.degen_tol,symm_tol=args.sym_tol,use_save_symmetries=args.use_save_symm)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function, division
import os
import re
from math import sqrt
import numpy as np
from .lattice import *
from qepy.auxiliary import *

eVtocm1 = 8065.54429
cm1toeV = 1.0/eVtocm1
ha2ev  = 27.211396132
eV2ha  = 1.0/ha2ev
Thz2cm1 = 33.35641
cm12Thz = 1.0/33.35641

__all__ = [
"Matdyn",
]

class Matdyn(object):
    """
    Class to read and plot the data from matdyn.modes files 
    """

    def __init__(self,qpoints,eig,eiv):
        # Notice that q-points are in cartesian units of 2pi/a_0
        self.qpoints  = np.array(qpoints)
        self.eig      = np.array(eig)
        self.eiv      = np.array(eiv)

    @classmethod
    def from_modes_file(cls,folder='.',filename='matdyn.modes'):
        """
        read the modes file from the hard drive 
        """
        with open(os.path.join(folder, filename),'r') as f:
            data_phon = f.readlines()

        #detect dimensions of the file
        #qpoints
        nqpoints = 0
        for line in data_phon:
            if 'q =' in line:
                nqpoints+=1
        nqpoints = nqpoints

        #modes
        nline = 0
        while 'freq' not in data_phon[nline]:
            nline += 1
        start_line = nline
        nline +=1
        while 'freq' not in data_phon[nline]:
            nline += 1
        end_line = nline
        natoms = end_line-start_line-1
        nmodes = natoms*3

        #empty stuff
        eig = []
        eiv = []
        qpoints = []

        #read qpoints, modes and energies
        for j in range(nqpoints):
            frec, v_frec = [], []
            k=2 + j*(nmodes*(natoms+1)+5)
            qpoints.append( float_from_string(data_phon[k]) )
            for i in range(nmodes):
                k=4 + j*(nmodes*(natoms+1)+5) + i*(natoms+1)
                y = float_from_string(data_phon[k])
                v_mode = []
                for ii in range(1,natoms+1):
                    z      = float_from_string(data_phon[k+ii])
                    v_atom = [complex(z[0],z[1]),complex(z[2],z[3]),complex(z[4],z[5])]
                    v_mode.append(v_atom)
                v_frec.append(v_mode)
                frec.append(y[1])
            eig.append(frec)
            eiv.append(v_frec)

        #store info
        natoms  = natoms
        nmodes  = nmodes
        nqpoints= nqpoints
        qpoints = np.array(qpoints)
        eig     = np.array(eig)
        eiv     = np.array(eiv).reshape(nqpoints,nmodes,nmodes)
        return cls(qpoints,eig,eiv)

    @property
    def modes(self):
        return self.eiv.reshape(self.nqpoints,self.nmodes,self.natoms,3)

    @property
    def nmodes(self):
        nqpoints,nmodes,_ = self.eiv.shape 
        return nmodes
 
    @property
    def natoms(self):
        nqpoints,nmodes,_ = self.eiv.shape 
        return int(nmodes/3)

    @property
    def nqpoints(self):
        nqpoints,nmodes,_ = self.eiv.shape 
        return nqpoints

    @property
    def write_modes(self,filename=None):
        """
        save the phonon modes in a file
        """
        s = " matrix written with qepy\n\n"
        for nq in range(self.nqpoints):
            s += ("q =  "+"%12.6lf "*3+"\n")%tuple(self.qpoints[nq])
            s += "*"*81+"\n"
            for n,mode in enumerate(self.eiv[nq]):
                phfreqmev = self.get_phonon_freq(nq,n+1,unit='THz')
                phfreqcm1 = self.get_phonon_freq(nq,n+1,unit='cm-1')
                s += '    freq ( %4d) = %12.6lf [ThZ] = %12.6lf [cm-1]\n'%(n+1,phfreqmev,phfreqcm1)
                for a in range(self.natoms):
                    xr,yr,zr = mode[a*3:(a+1)*3].real
                    xi,yi,zi = mode[a*3:(a+1)*3].imag
                    s += ("( "+"%12.6lf "*6+')\n')%(xr,xi,yr,yi,zr,zi)
            s += "*"*81+"\n"
            s += "\n\n"
        
        if filename:
            f = open(filename,'w')
            f.write(s)
            f.close()
        else:
            print(s)

    def rotate_phonons(self,eps=1e-5,debug=False):
        """
        Rotate the degenerate states to align then with the x,y axis
        Arguments:
        eps -> threshold for finding degeneracies
        """
        eig = self.eig
        eiv = self.eiv
        
        #the eigenvalues are probably sorted but just in case...
        eig, eiv = zip(*sorted(zip(eig,eiv), key=lambda x: x[0]))
        eig = np.array(eig)
        eiv = np.array(eiv)
        
        #iterate over qpoints
        for nq,(eigq,eivq) in enumerate(zip(eig,eiv)):
            #detect the degeneracies
            keys = [group.mean() for group in np.split(eigq, np.where(np.diff(eigq) > eps)[0]+1)]
            unique_frequencies = dict([(key,0) for key in keys])
            #count degeneracies
            for eig in eigq:
                for key in keys:
                    if np.isclose(key,eig,atol=eps):
                        unique_frequencies[key] += 1
            for val,deg in sorted(unique_frequencies.items()):
                if debug: print("frequency: %12.8lf [Thz] %12.8lf [cm-1] degeneracy: %2d" % (val,val,deg))
                if deg > 1:
                    #get the indexes of the modes with the same frequency
                    indexes = [n for n,f in enumerate(eigq) if np.isclose(f,val,atol=eps)]
                    r = np.array(eivq[indexes])
                    if debug:
                        print("input basis:")
                        for n in range(deg):
                            print("mode: %3d"%n)
                            for i in range(self.natoms):
                               print("atom %3d"%i+("%12.8lf"*3)%tuple(r[n,i*3:(i+1)*3].real))
                    #we make sure the first column vector the matrix r in non zero
                    rows,cols = r.shape
                    n = 0
                    while np.isclose(r[:,n],np.zeros([len(r)])).all():
                        n += 1
                    q, a = np.linalg.qr(r[:,n:])
                    r[:,n:] = a
                    if debug:
                        print("canonical basis:")
                        for n in range(deg):
                            print("mode: %3d"%n)
                            for i in range(self.natoms):
                                print("atom %3d"%i+("%12.8lf"*3)%tuple(r[n,i*3:(i+1)*3].real))
                    eivq[indexes] = r
                self.eiv[nq] = eivq

    def plot_eigen(self,path=[]):
        """ plot the phonon frequencies using matplotlib
        """
        import matplotlib.pyplot as plt

        if path:
            if isinstance(path,Path):
                path = path.get_indexes()
            plt.xticks( *list(zip(*path)) )
        plt.ylabel('\\omega (cm$^{-1}$)')

        #plot vertical line
        for point in path:
            x, label = point
            plt.axvline(x)

        #plot bands
        eig = np.array(self.eig)
        for ib in range(self.nmodes):
           plt.plot(range(self.nqpoints),eig[:,ib], 'r-', lw=2)
        plt.show()

    def get_phonon_freq(self,iq,n,unit="eV"):
        """
        Get the value of the phonon frequency
        nq -> q-point from where to get the frequency from
        n  -> mode of the phonon
        """
        if   unit == "eV":
            factor = cm1toeV
        elif unit == "Ha":
            factor = cm1toeV*eV2ha
        elif unit == "THz":
            factor = cm12Thz
        elif unit == "cm-1":
            factor = 1
        else:
            raise ValueError('Unit %s not known'%unit)

        if n==0:
            raise ValueError('Error! Phonon modes start from 1')

        return self.eig[iq][n-1]*factor

    def normalize(self):
        """
        Normalize the displacements u^n_{ai} according to:
        
        sum_ai ( u^n_{ai} )**2 = 1
        """
       
        for nq in range(self.nqpoints):
            for n in range(self.nmodes):
                print(np.linalg.norm(self.eiv[nq,n]))
                self.eiv[nq,n] /= np.linalg.norm(self.eiv[nq,n])

    def normalize_with_masses(self,masses): 
        """
        Normalize the displacements u^n_{ai} according to:
        
        sum_{ai} M_a u^n_{ai} u^m_{ai} = delta_{nm}
        
        u -> displacement
        n -> phonon mode
        a -> atom index
        i -> direction
        M -> mass
        """

        masses = np.array(masses)
        ref_mass = max(masses)
        masses = masses/ref_mass

        #divide by masses
        if self.check_orthogonality():
            for nq in range(self.nqpoints):
                for n in range(self.nmodes):
                    for a in range(self.natoms):
                        self.eiv[nq,n,a*3:(a+1)*3] *= 1.0/sqrt(masses[a])
        else:
            print("These eigenvectors are non-orthogonal, probably they are already scaled by the masses so I won't do it")

        #enforce delta_nm
        for nq in range(self.nqpoints):
            for n in range(self.nmodes):
                s = 0
                for a in range(self.natoms):
                    e = self.eiv[nq,n,a*3:(a+1)*3]
                    #get normalization constant
                    s += masses[a]*np.vdot(e,e).real
                self.eiv[nq,n] *= 1.0/sqrt(s)
    
    def check_orthogonality(self,atol=1e-5):
        """
        Check if the eigenvectors are orthogonal
        """

        orth = np.zeros([self.nmodes,self.nmodes])
        for nq in range(self.nqpoints):
            for n in range(self.nmodes):
                e1 = self.eiv[nq,n]
                for m in range(self.nmodes):
                    e2 = self.eiv[nq,m]
                    orth[n,m] = np.vdot(e1,e2).real
        
        return np.isclose(orth,np.eye(self.nmodes),atol=atol).all()
 
    def check_normalization(self,masses,atol=1e-5):
        """
        Check if the displacements are normalized according to:
        
        sum_{ai} M_a u^n_{ai} u^m_{ai} = delta_{nm}
        """
        
        masses = np.array(masses)
        ref_mass = max(masses)
        masses = masses/ref_mass
        
        #check normalization
        norm = np.zeros([self.nmodes])
        for nq in range(self.nqpoints):
            for n in range(self.nmodes):
                s = 0
                for a in range(self.natoms):
                    e = self.eiv[nq,n,a*3:(a+1)*3]
                    #get normalization constant
                    s += masses[a]*np.vdot(e,e).real
                norm[n] = s

        return np.isclose(norm,np.ones(self.nmodes),atol=atol).all()
 
    def __str__(self):
        s = ""
        for nq in range(self.nqpoints):
            for n,mode in enumerate(self.eiv[nq]):
                phfreqmev = self.get_phonon_freq(nq,n+1,unit='eV')*1000
                phfreqcm1 = self.get_phonon_freq(nq,n+1,unit='cm-1')
                s+= 'mode: %d freq: %8.2lf meV %8.2lf cm-1\n'%(n+1,phfreqmev,phfreqcm1)
                for a in range(self.natoms):
                    s += ("%12.8lf "*3+'\n')%tuple(mode[a*3:(a+1)*3].real)
        return s



# Copyright (c) 2018, Henrique Miranda
# All rights reserved.
#
# This file is part of the yambopy project
#
from yambopy import *
from yamboparser import *
import os

class YamboHFDB():
    """
    Class to read yambo ndb.HF_and_locXC files
    """
    def __init__(self,filename='ndb.HF_and_locXC',folder='.'):
        """
        Read a QP file using the yamboparser
        """
        self.folder = folder
        self.filename = filename
        if os.path.isfile('%s/%s'%(folder,filename)):
            self.yfile = YamboFile(filename,folder)
        else:
            raise ValueError('File %s/%s not found'%(folder,filename))

        qps = self.yfile.data
        nkpoints = qps['nkpoints']
        nbands = qps['nbands']
        self.sx = qps['Sx'].reshape(nkpoints,nbands)
        self.vxc = qps['Vxc'].reshape(nkpoints,nbands)

    def __str__(self):
        s = ""
        s += "nqps:     %d\n"%self.nqps
        s += "nkpoints: %d\n"%self.nkpoints
        s += "nbands:   %d\n"%self.nbands
        return s




#
# 1st version by Fulvio Paleari
# This file is part of yambopy
#
import os
import re
import numpy as np
from qepy import PwIn
from yambopy.lattice import rec_lat, car_red,red_car
from yambopy.zeros import matdyn_q_atol 
from itertools import product
import copy
from math import *

# Dimensional constants for reference (actually only b2a is really used in the code for now)
cm1_2_Tera=0.0299793         # Conversion from cm-1 to THz with 2pi factor included
Tera=1.e12                   
b2a =0.529177                
hbar=6.5821e-16              # Planck's constant (eV*s)
kb=8.6173e-5                 # Boltzmann's constant (eV/K)
Mp=1.0073                    # Proton mass (reference, u)
cMp=Mp*1.660539*6.241509e-29 # Conversion of Mp in eV*\AA^{-2}*s^2
#
## ISSUES TO FIX ##
"""(iii) Small issue in nondiagonal supercell matrices for certain q-vectors
   (iv)  In nondiagonal supercells: the suggested kpoint mesh must be CONSISTENT and UNIFORM --> to fix
"""

## Start of the proper supercell class
class Supercell():
    """
        A class to generate custom supercells from a quantum espresso input file.
        The supercell may be "non-diagonal" according to [Lloyd-Williams and Monserrat, Phys. Rev. B 92, 184301, 2015].
        The atoms in the supercell may also be displaced along phonon modes if provided.
        
        Input arguments:
       
            - qe_input: a PwIn() instance of an input file in the unit cell (uc)
            - Optional: QE-DFPT phonon modes output
        
        How it works:
            
            - Call self.d_sup(R) to build diagonal supercell
                -- R is a list of repetitions of the uc in the cartesian directions
            - Call self.nd_sup(Q) to build nondiagonal supercell
                -- Q contains the fractional coordinates of the q-point to be folded at Gamma in a nondiagonal supercell like [[m1,m2,m3],[n1,n2,n3]]
            - Call self.displace(modes_file,new_atoms,Temp=0.1) to displace supercell
                -- modes_file is a QE-DFPT phonon-mode output
                -- Temp is the width of the displacements in bohr
                -- new_atoms is the output of (n)d_sup
                
        Sample input script found at:
            tutorials/supercell
            
        NOTA BENE:
            The Q-vector to be folded must be given in CRYSTAL COORDINATES. If computing its related phonons,
            remember that the output of quantum espresso uses CARTESIAN COORDINATES
    """

    def __init__(self,qe_input):

        self.qe_input = qe_input
        self.latvec   = np.array(qe_input.cell_parameters)
        self.basis    = int(qe_input.system['nat'])
        self.atoms    = qe_input.atoms
        self.uc_kpts  = qe_input.kpoints
        self.atypes   = qe_input.atypes
        self.aunits   = qe_input.atomic_pos_type
        #Atomic masses (in u)
        self.m_at = np.array( [ float( self.atypes.get( self.atoms[i][0] )[0] ) for i in range(self.basis) ] )


    """
    [START] Displacement-related functions                          
    """
    def displace(self,qe_dyn,new_atoms,iq=0,Temp=0.1,use_temp='no',write=True):
        """
        Case of displaced supercell
        """
        #Check if we are displacing the unit cell (i.e., gamma modes)
        GAMMA = False
        try: self.Q
        except AttributeError: GAMMA = True
        
        print('Applying displacements according to phonon modes...')
        self.use_temp = use_temp
        self.initialize_phonons(iq,qe_dyn,Temp)

        if GAMMA and np.linalg.norm(qe_dyn.qpoints[iq])>matdyn_q_atol:
            print("WARNING: Q-point in matdyn file different from the one used to generate supercell")
        if not GAMMA:
            # Transform all q-point in reduced coordinates
            rlat =rec_lat(self.latvec)
            alat0=self.qe_input.get_alat0()
            q = []
            q.append(qe_dyn.qpoints[iq]/alat0)
            q_matdyn=car_red(q,rlat)
            q_in = np.array([float(self.Q[0,i])/float(self.Q[1,i]) for i in range(3)])

            # Bring in the BZ [0,1)
            q_in = q_in-np.floor(q_in)
            q_matdyn = q_matdyn-np.floor(q_matdyn)
            if not np.allclose(q_matdyn,q_in,rtol=0.0,atol=matdyn_q_atol):
                print("WARNING: Q-point in matdyn file different from the one used to generate supercell")
                print("Q-in     : ",q_in, " [red] ")
                print("Q-matdyn : ",q_matdyn[0], " [red] ")

        if GAMMA: #No phases and take only optical modes
            phases = np.ones(self.sup_size)
            expand_eigs = np.array([phases[i]*self.eiv for i in range(self.sup_size)])
            self.print_expanded_eigs(expand_eigs,GAMMA=GAMMA)
        else: 
            phases = self.getPhases()                
            expand_eigs = np.array([phases[i]*self.eiv for i in range(self.sup_size)])
            self.print_expanded_eigs(expand_eigs,GAMMA=GAMMA) #Print expanded eigs
            #Take real part
            for cell in range(self.sup_size): expand_eigs[cell]= self.take_real(expand_eigs[cell])            

        disps = expand_eigs.real.astype(float)
        #Force same gauge choice
        #for cell in range(self.sup_size): disps[cell]= self.force_gauge(disps[cell])
        #Transform eigenmodes in displacements
        #disps[cell][mode][basis][direction]
        disps = np.array([self.osc_length(disp_slice,GAMMA=GAMMA) for disp_slice in disps])
        #disps[mode][cell][basis][direction]
        self.disps = disps.swapaxes(0,1)
        if GAMMA: self.disps = self.disps[3:]    
        if write:
            #A list of PwIn() objects (one for each phonon mode) that can be printed, written to file, etc.
            if 'Q' in globals(): mode='nd'
            else: mode='diagonal'
            self.modes_qe = [self.write(new_atoms,mode,phonon=disps_slice) for disps_slice in self.disps]

 
    def print_expanded_eigs(self,exp_eigs,GAMMA=False):
        """
        Print expanded modes in QE-style, to compare and for reference
        """
        if GAMMA:
            q = np.array([0.,0.,0.])
            ncells = 1
            mode_start = 3
        else:
            q = np.array([float(self.Q[0,i])/float(self.Q[1,i]) for i in range(3)])
            ncells = int(np.max(np.array(self.Q[1])))
            mode_start = 0
        
        sc_basis = self.basis*ncells
        freq_THz =[self.qe_dyn.get_phonon_freq(self.iq,im+1,unit='THz')  for im in range(self.qe_dyn.nmodes) ]
        freq_cmm1=[self.qe_dyn.get_phonon_freq(self.iq,im+1,unit='cm-1') for im in range(self.qe_dyn.nmodes) ]
        exp_eigs = exp_eigs.swapaxes(0,1) #[mode][cell][basis][direction]
        if GAMMA: filename = self.qe_input.control['prefix'][1:-1]+"_s.modes_GAMMA"
        else: filename = self.qe_input.control['prefix'][1:-1]+"_s.modes_expanded"
        exp_eigs_file = open(filename,"w")
        exp_eigs_file.write("     diagonalizing the dynamical matrix ...\n")
        exp_eigs_file.write("\n")
        exp_eigs_file.write(" q =     %2.5f   %2.5f   %2.5f\n"%(q[0],q[1],q[2]))
        exp_eigs_file.write(" **************************************************************************\n") 
        for m in range(mode_start,3*self.basis):
            exp_eigs_file.write("     freq (   %d) =       %2.6f [THz] =       %2.6f [cm-1]\n" \
                                %(m+1,freq_THz[m],freq_cmm1[m]))
            for c in range(ncells):
                for a in range(self.basis):
                    exp_eigs_file.write("( %2.6f  %2.6f  %2.6f  %2.6f  %2.6f  %2.6f )\n" \
                                        %(exp_eigs[m,c,a,0].real,exp_eigs[m,c,a,0].imag, \
                                          exp_eigs[m,c,a,1].real,exp_eigs[m,c,a,1].imag, \
                                          exp_eigs[m,c,a,2].real,exp_eigs[m,c,a,2].imag))
        exp_eigs_file.write(" **************************************************************************")
        exp_eigs_file.close()

    
    def initialize_phonons(self,iq,qe_dyn,Temp):
        self.Temp     = Temp
        self.qe_dyn   = qe_dyn
        self.iq       = iq
        self.Omega    = qe_dyn.eig[iq]*Tera
        self.eiv = np.reshape(self.qe_dyn.eiv[iq],(self.qe_dyn.nmodes,self.basis,3))

    def getPhases(self):
        q = np.array([float(self.Q[0,i])/float(self.Q[1,i]) for i in range(3)])
        arg = q[0]*self.T[:,0]+q[1]*self.T[:,1]+q[2]*self.T[:,2]
        return np.exp(1j*2.*np.pi*arg)

    def take_real(self,eig):
        modes=3*self.basis
        for i in range(modes):
            #Check that there aren't purely imaginary modes
            if np.real(eig[i]).all == 0.:
                print("mode %i: Purely imaginary"%(i+1))
                eig[i]=1j*eig[i]
        eig = np.real(eig)
        return eig

    def force_gauge(self,eig):
        """ 
        for each normal mode, the first nonzero element is set to be positive
        see page 075125-3 of PRB 94, 075125 (2006)
        """
        modes=3*self.basis
        for i in range(modes):
            a=np.nonzero(eig[i])                 # for each normal mode
            a=np.array(a)                        # I get the nonzero elements, of which the first
            if eig[i,a[0,0],a[1,0]] < 0.:        #one, given by these indices, must be >0
                eig[i]=-eig[i]
        return eig

    def osc_length(self,eig,GAMMA=False):
        """
        Oscillator lengths per mode (in ANGSTROM)
        NB:  Only ARBITRARY displacements can be set.
        NB2: Eigenmodes from quantum espresso are ALREADY weighted by atomic masses
        """
        RESCALE     = b2a*self.Temp  #Arbitrary displacement
        temperature = 0.*self.Temp      #Harmonic displacement (permanently set to zero for now)
        modes=3*self.basis
        displacements=[]
        lengths_per_mode=[]
        for nu,eig_slice in enumerate(eig):
            
            if GAMMA and ( nu==0 or nu==1 or nu==2): l=0.
            else: l=sqrt(hbar/(2.*cMp*self.Omega[nu]))
            
            if temperature==0.: sigma2=l*l
            else: sigma2=l*l*(2./(np.exp(hbar*self.Omega[nu]/(kb*temperature))-1.)+1)
            
            if self.use_temp=='no': displacements.append(RESCALE*eig_slice)     #Each mode (i.e. atomic displacement directions) is multiplied by the corresponding length
            else:                   displacements.append(np.sqrt(sigma2)*eig_slice) #Displacement by harmonic sigma
            #List of average "realistic" atomic displacements (not weighted by atomic mass)
            lengths_per_mode.append(np.sqrt(sigma2))
        displacements = np.array(displacements)
        lengths_per_mode = np.array(lengths_per_mode)
        #mass_ratio = np.array([sqrt(Mp/mass) for mass in self.m_at])
        #for d,i in product(range(len(displacements)),range(self.basis)): displacements[d,i,:] *= mass_ratio[i] #Weigh the displ. with the different masses
        
        return displacements #displacements[mode in order of ascending frequency][basis]

    """
    [END] Displacement-related functions   
    """
    
    def lattice_constants(self,vec):
        return [np.linalg.norm(vec[0]),np.linalg.norm(vec[1]),np.linalg.norm(vec[2])]

    def d_sup(self,R,write=True):
        """
        Case of diagonal supercell
        """     
        self.R = R
        self.sup_size = int(R[0]*R[1]*R[2])
        self.new_latvec = np.array([self.latvec[i]*R[i] for i in range(3)])
        self.sup_size = int(self.R[0]*self.R[1]*self.R[2])
        new_atoms = self.build_supercell()
        if write: 
            #PwIn() object that can be printed, written to file, etc.
            self.qe_d = self.write(new_atoms,mode='diagonal')
        return new_atoms

    def nd_sup(self,Q,write=True):
        """
        Case of nondiagonal supercell
        """
        self.Q = np.array(Q)
        print('Nondiagonal supercell')
        if (self.uc_kpts % self.Q[1] != 0).any():
            print('ERROR: You must set a unit cell k-point mesh where%s Nx,Ny,Nz are multiples of %d,%d,%d, respectively.'%('\n',self.Q[1,0],self.Q[1,1],self.Q[1,2]))
            exit()
        self.R, self.new_latvec = self.find_nondiagonal()
        self.sup_size = int(self.R[0]*self.R[1]*self.R[2])
        new_atoms = self.build_supercell()
        if write:
            self.qe_nd = self.write(new_atoms,mode='nd')
        return new_atoms

    def build_supercell(self):
        latvec     = self.latvec
        R          = self.R
        atoms      = np.array([atom[1] for atom in self.atoms])
        if self.aunits!='angstrom': atoms = red_car(atoms,latvec) 
        else: latvec = b2a*latvec
        #new_atoms[cell][basis][direction]
        new_atoms      = np.array([atoms for n in range(self.sup_size)])
        T = []
        for nz,ny,nx in product(range(int(R[2])),range(int(R[1])),range(int(R[0])) ): 
            cell=int(nx+ny*R[0]+nz*R[0]*R[1]) 
            translation = nx*latvec[0] +ny*latvec[1] +nz*latvec[2]
            for b in range(self.basis): 
                new_atoms[cell,b]=new_atoms[cell,b] + translation 
            T.append(translation)
        T = np.array(T) #Positions of the repeated unit cells
        if self.aunits!='angstrom': self.T=car_red(T,self.latvec)
        else: self.T=T
        #new_atoms[super_basis][directions]$
        new_atoms=new_atoms.reshape(self.basis*self.sup_size,3) 
        if self.aunits!='angstrom': new_atoms = car_red(new_atoms,self.new_latvec)
        return new_atoms

    def find_integers(self,nums,g23,g12,g31,g123):
        """
        Compute integers for off-diagonal supercell matrix elements 
        Called by find_nondiagonal()
        """
        if nums[1]==0: p=0
        else: 
            #Compute p (it's a modulo equation)
            if g23 == 1: p = 0
            else:
                for i in range(1,g23):
                    if (nums[1]+i*nums[2]) % g23 == 0:
                        p=i
                        break
        if nums[0]==0: q,r=[0,1] #[POSSIBLE BUG for certain q-vectors] These conditions must be checked carefully
        else:
            #Compute q
            g12_r = int(g12/g123)
            g23_r = int(g23/g123)
            g31_r = int(g31/g123)
            if g12_r == 1: q = 0
            else:
                for i in range(1,g12_r):
                    if (g23_r*nums[0]+i*g31_r*nums[1]) % g12_r == 0:
                        q=i
                        break
            #Compute r
            gg_r = g31*g23/g123
            z = g23*nums[0]/g12+g31*q*nums[1]/g12
            if gg_r == 1: r = 0
            else:
                for i in range(1,round(gg_r)):
                    if (z+i*nums[2]) % gg_r == 0:
                        r=i
                        break
        return p,q,r 
 
    def find_nondiagonal(self):
        """
        Nondiagonal supercell, based on [Phys. Rev. B 92, 184301]
        """
        Q = self.Q
        #Take care of components already at Gamma
        Q[1,np.where(Q[0]==0)]=1
        #Shift the q-point into the positive quadrant of the reciprocal unit cell
        Q[0,np.where(Q[0]<0)]+=Q[1,np.where(Q[0]<0)]
        #GCDs of Q[1] (in the logical order of the derivation)
        g23  = gcd(Q[1,1],Q[1,2])
        g12  = gcd(Q[1,0],Q[1,1])
        g31  = gcd(Q[1,2],Q[1,0])
        g123 = gcd(Q[1,0],gcd(Q[1,1],Q[1,2]))
        #Integers needed to solve the supercell matrix equation    
        p,q,r = self.find_integers(Q[0],g23,g12,g31,g123)            
        #Matrix elements (in order of derivation) and supercell matrix
        S_33 =        Q[1,2]
        S_22 =        Q[1,1]/g23
        S_23 =      p*Q[1,2]/g23
        S_11 =   g123*Q[1,0]/(g12*g31)
        S_12 = q*g123*Q[1,1]/(g12*g23)
        S_13 = r*g123*Q[1,2]/(g31*g23)
        self.S = np.array([[S_11,S_12,S_13],[0,S_22,S_23],[0,0,S_33]])
        #New lattice vectors and actual supercell size
        new_latvec = np.einsum('ij,jx->ix',self.S,self.latvec)
        R          = [self.S[0,0],self.S[1,1],self.S[2,2]]
        print(self.S)
        return R, new_latvec

    def reciprocal(self,mode):
        """
        Function to compute reciprocal lattice
        """
        #Unit cell
        repvec = rec_lat(self.latvec)
        alat=np.array(self.lattice_constants(self.latvec))
        self.repvec = 2.*np.pi*np.multiply(1./alat,repvec)
        #Supercell
        if mode=='diagonal': self.new_repvec = np.array([self.repvec[i]/float(R[i]) for i in range(3)])
        else: 
            self.S_inv_T = np.linalg.inv(self.S).T
            self.new_repvec = np.einsum('ij,jx->ix',self.S_inv_T,self.repvec)

    def atoms_input(self, new_atoms):
        """ 
        Put the atomic element labels in the right order
        """
        positions_input = new_atoms.tolist()
        elements_input  = [[self.qe_input.atoms[i][0] for i in range(int(self.basis))] for j in range(self.sup_size)]  
        elements_input  = [ item for sublist in elements_input for item in sublist ]
        atoms_input     = [[elements_input[i], positions_input[i]] for i in range(self.sup_size*self.basis)]
        return atoms_input

    def posint(self,value):
        return abs(int(round(value)))

    def write(self,new_atoms,mode,phonon=None):
        R = self.R
        new_latvec = self.new_latvec
        alat = self.lattice_constants(new_latvec)
        qe = self.qe_input
        if mode=='diagonal':
            #A suggestion for a consistent new kpoint mesh 
            new_kpoints = [ceil(qe.kpoints[0]/R[0]), ceil(qe.kpoints[1]/R[1]), ceil(qe.kpoints[2]/R[2])]
        else:
            #The compulsory new kpoint mesh - (sub)multiples of it are also fine but not consistent
            self.reciprocal('nondiagonal')
            new_kpoints = np.dot(self.S_inv_T,np.array(qe.kpoints))
            new_kpoints = [self.posint(new_kpoints[0]),self.posint(new_kpoints[1]),self.posint(new_kpoints[2])]
        if phonon is not None:
            phonon = phonon.reshape(self.basis*self.sup_size,3)
            phonon = car_red(phonon,self.new_latvec)
            new_atoms = new_atoms + phonon
        qe_s = copy.deepcopy(qe)
        qe_s.set_atoms(self.atoms_input(new_atoms))
        qe_s.control['prefix'] = qe.control['prefix'][:-1]+"_s'"
        #[POSSIBLE BUG] with only ibrav==0 and cell_parameters, it might fail the symmetry !?
        if 'celldm(1)' in qe_s.system: del qe_s.system['celldm(1)']
        if 'celldm(2)' in qe_s.system: del qe_s.system['celldm(2)']
        if 'celldm(3)' in qe_s.system: del qe_s.system['celldm(3)']
        """
        qe_s.system['celldm(1)'] = alat[0]
        qe_s.system['celldm(2)'] = alat[1]/alat[0]
        qe_s.system['celldm(3)'] = alat[2]/alat[0]
        """
        qe_s.system['ibrav']=0
        qe_s.cell_units = 'bohr'
        qe_s.cell_parameters = new_latvec
        #Just a suggestion for the new bands
        if 'nbnd' in qe.system: qe_s.system['nbnd'] = self.sup_size*int(qe.system['nbnd'])
        qe_s.system['nat'] = self.basis*self.sup_size
        qe_s.kpoints = new_kpoints
        return qe_s





from yambopy import *
from yambocommandline import *
import os
import glob

# 1. pack files of convergence GW calculations
# important: all calculations must finish
pack_files_in_folder('bse_conv')

# 2. Read json files
ya = YamboAnalyser('bse_conv')

# Plot BSE spectra for each parameter
ya.plot_bse(('eps_q1','FFTGvecs'),cols=(2,),png_file=True)

ya.plot_bse(('eps_q1','NGsBlkXs'),cols=(2,),png_file=True)

ya.plot_bse(('eps_q1','BndsRnXs'),cols=(2,),png_file=True)

if glob.glob('bse_conv/BSENGBlk*'):
    ya.plot_bse(('eps_q1','BSENGBlk'),cols=(2,),png_file=True)

if glob.glob('bse_conv/BSENGexx*'):
    ya.plot_bse(('eps_q1','BSENGexx'),cols=(2,),png_file=True)

if glob.glob('bse_conv/BSEEhEny*'):
    ya.plot_bse(('eps_q1','BSEEhEny'),cols=(2,),png_file=True)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP, HPC, AMS
#
# This file is part of the yambopy project
#
import os
import numpy as np
from netCDF4 import Dataset
from yambopy.tools.string import marquee
from yambopy.tools.funcs import fermi, fermi_array
from yambopy.lattice import car_red, rec_lat, vol_lat
from yambopy.kpoints import expand_kpoints, get_path
from yambopy.plot.spectra import get_spectra
from yambopy.units import ha2ev

class YamboElectronsDB():
    """
    Class to read information about the electrons from the ``ns.db1`` produced by yambo

    Arguments:

        ``filename``: netcdf database to read from (default:ns.db1)

    NB: - spin polarized calculations not yet supported
        - spin-polarized eigenvalues are read and expanded for compatibility with DipolesDB
    """

    def __init__(self,atomic_numbers,car_atomic_positions,eigenvalues_ibz,sym_car,
                 iku_kpoints,nbands,lat,alat,temperature,nelectrons,nkpoints_ibz,spin,time_rev,spinor_components,Expand):
        
        self.atomic_numbers         = atomic_numbers   
        self.car_atomic_positions   = car_atomic_positions
        self.eigenvalues_ibz        = eigenvalues_ibz     
        self.sym_car                = sym_car        
        self.iku_kpoints            = iku_kpoints
        self.nbands                 = nbands       
        self.lat                    = lat             
        self.alat                   = alat            
        self.temperature            = temperature     
        self.nelectrons             = nelectrons
        self.nkpoints_ibz           = nkpoints_ibz       
        self.spin                   = spin            
        self.time_rev               = time_rev
        self.spinor_components      = spinor_components
        self.EXPAND                 = Expand # expand eigenvalues and kpoints to full BZ
        self.expanded               = False  # flag to check if the kpoints were already expanded (starts from False)

        if self.EXPAND: self.expandEigenvalues()

    @classmethod
    def from_db_file(cls,folder='.',filename='ns.db1',Expand=False):
        """
        Read the ns.db1 database
        """
        path_filename = os.path.join(folder,filename)
        try:    database = Dataset(path_filename)
        except: raise IOError("Error opening file {path_filename} in YamboElectronsDB")

        with Dataset(path_filename) as database:
            
            dimensions            = database.variables['DIMENSIONS'][:]
            
            natoms_a = database.variables['N_ATOMS'][:].astype(int).T
            tmp_an = database.variables['atomic_numbers'][:].astype(int)
            tmp_apos = database.variables['ATOM_POS'][:,:]

            flatten = lambda l: [item for sublist in l for item in sublist]
            atomic_numbers = flatten([[tmp_an[n]]*na for n,na in enumerate(natoms_a)])
            atomic_positions = np.vstack([[tmp_apos[n,ia] for ia in range(na)] for n,na in enumerate(natoms_a) ])

            args = dict( atomic_numbers         = atomic_numbers,
                         car_atomic_positions   = atomic_positions,
                        eigenvalues_ibz        = database.variables['EIGENVALUES'][...].data*ha2ev,
                         sym_car                = np.transpose( database.variables['SYMMETRY'][:], (0,2,1) ), # transpose leaving first axis as symm index
                         iku_kpoints            = database.variables['K-POINTS'][:].T,
                         nbands                 = int(dimensions[5]),
                         lat                    = database.variables['LATTICE_VECTORS'][:].T,
                         alat                   = database.variables['LATTICE_PARAMETER'][:].T,
                         temperature            = dimensions[13],
                         nelectrons             = int(dimensions[14]),
                         nkpoints_ibz           = int(dimensions[6]),
                         spin                   = int(dimensions[12]),
                         time_rev               = dimensions[9],
                         spinor_components      = int(dimensions[11]),
                         Expand                 = Expand 
                         )

        return cls(**args)

    @property
    def red_atomic_positions(self):
        return car_red(self.car_atomic_positions,self.lat)

    @property
    def spin_degen(self):
        """spin degeneracy if 2 components degen 1 else degen 2"""
        return [0,2,1][int(self.spin)]

    @property
    def min_eival(self):
        return np.min(self.eigenvalues) 
    
    @property
    def max_eival(self):
        return np.max(self.eigenvalues)

    @property
    def car_kpoints(self):
        """convert form internal yambo units to cartesian lattice units"""
        return np.array([ k/self.alat for k in self.iku_kpoints ])

    @property
    def red_kpoints(self):
        """convert from cartesian coordinates to reduced coordinates"""
        if not hasattr(self,"_red_kpoints"):
            self._red_kpoints = car_red(self.car_kpoints,self.rlat)
        return self._red_kpoints

    @property
    def rlat(self):
        """caclulate the reciprocal lattice"""
        return rec_lat(self.lat)

    @property
    def rlat_vol(self):
        return (2*np.pi)**3 * vol_lat(self.rlat)

    @property
    def lat_vol(self):
        return vol_lat(self.lat)

    @property
    def natoms(self):
        return len(self.atomic_positions)

    @property
    def nbandsv(self):
        """
        number of occupied bands
        """
        if (self.spinor_components==2): return int(self.nelectrons) 
        else:                           return int(self.nelectrons/2)

    @property
    def nbandsc(self):
        return int(self.nbands-self.nbandsv)

    @property
    def nbands_tot(self):
        """
        NB: in the spin-polarised case, nbands contains the total number
            of bands PER spin polarisation, i.e. half of the total number.
            Therefore, nbandsv and nbandsc are also given per
            spin polarisation: this fact is used by DipolesDB    
        """
        if self.spin==2: return self.nbands*self.spin
        else           : return self.nbands
    
    @property
    def nbandsv_tot(self):
        if self.spin==2: return int(self.nelectrons/self.spin_degen)
        else:            return self.nbandsv
    
    @property
    def nbandsc_tot(self):
        if self.spin==2: int(self.nbands_tot-self.nbandsv_tot)
        else:            return self.nbandsc

    @property
    def time_rev_list(self):
        """get a list of symmetries with time reversal"""
        time_rev_list = [False]*self.nsym
        for i in range(self.nsym):
            time_rev_list[i] = ( i >= self.nsym/(self.time_rev+1) )
        return time_rev_list

    @property
    def sym_rlu(self):
        """convert cartesian transformations to reduced transformations """
        sym_rlu = np.zeros([self.nsym,3,3])
        for n,s in enumerate(self.sym_car):
            a = np.dot(s.T,np.linalg.inv(self.rlat))
            sym_rlu[n] = np.dot(np.linalg.inv(self.lat.T),a)
        return sym_rlu

    @property
    def nsym(self):
        return len(self.sym_car)

    @property
    def sym_red(self):
        """Convert cartesian transformations to reduced transformations"""
        if not hasattr(self,"_sym_red"):
            sym_red = np.zeros([self.nsym,3,3],dtype=int)
            for n,s in enumerate(self.sym_car):
                sym_red[n] = np.round(np.dot(np.dot(self.lat,s.T),np.linalg.inv(self.lat)))
            self._sym_red = sym_red
        return self._sym_red

    @property
    def sym_rec_red(self):
        """Convert reduced transformations to reduced reciprocal transformations"""
        if not hasattr(self,"_sym_rec_red"):
            sym_rec_red = np.zeros([self.nsym,3,3],dtype=int)
            for n,s in enumerate(self.sym_red):
                sym_rec_red[n] = np.linalg.inv(s).T
            self._sym_rec_red = sym_rec_red
        return self._sym_rec_red

    @property
    def sym_rec(self):
        """Convert cartesian transformations to reciprocal transformations"""
        sym_rec = np.zeros([self.nsym,3,3])
        for n,s in enumerate(self.sym_car):
            sym_rec[n] = np.linalg.inv(s).T
        return sym_rec

    @property
    def efermi(self):
        if not hasattr(self,"_efermi") and self.Expand:
            # AMS: To be tested
            self._efermi = self.GetFermi()
        return self._efermi

    def GetFermi(self,inv_smear=0.001,verbose=0,setfermi=True):
        """ Determine the fermi energy
        """
        from scipy.optimize import bisect

        kpts, nks, nss = self.expand_kpoints()

        def fermi_array(e_array,ef):
            """ Fermi dirac function for an array
            """
            e_array = (e_array-ef)/inv_smear
            return [ fermi(e) for e in e_array]

        def occupation_minus_ne(ef):
            """ The total occupation minus the total number of electrons
            """
            if self.spinor == 1:
               return sum([sum(self.spin_degen*fermi_array(self.eigenvalues[0,nk],ef))*self.weights[nk] for nk in range(self.nkpoints)])-self.electrons
            elif self.spinor == 2:
               sum_up = sum([sum(self.spin_degen*fermi_array(self.eigenvalues[0,nk],ef))*self.weights[nk] for nk in range(self.nkpoints)]) 
               sum_dw = sum([sum(self.spin_degen*fermi_array(self.eigenvalues[1,nk],ef))*self.weights[nk] for nk in range(self.nkpoints)]) 
               return sum_up + sum_dw -self.electrons
     
        efermi = bisect(occupation_minus_ne,self.min_eival,self.max_eival)

        if verbose: print("fermi: %lf eV"%efermi)

        if setfermi: return self.setFermi(efermi, inv_smear)
        else:        return efermi

    def setFermi(self,fermi,invsmear):
        """
        Shift bands and get occupations
        """
        self.invsmear = invsmear
        self.efermi = fermi

        #full brillouin zone
        self.eigenvalues     -= self.efermi
        self.occupations = np.zeros([self.spin,self.nkpoints,self.nbands],dtype=np.float32)
        for nspin in range(self.spin):
            for nk in range(self.nkpoints):
                self.occupations[nspin,nk] = fermi_array(self.eigenvalues[nspin,nk,:self.nbands],0)

        #for the ibz
        self.eigenvalues_ibz -= self.efermi
        self.occupations_ibz = np.zeros([self.spin,self.nkpoints_ibz,self.nbands],dtype=np.float32)
        for nspin in range(self.spin):
            for nk in range(self.nkpoints_ibz):
                self.occupations_ibz[nk] = fermi_array(self.eigenvalues_ibz[nspin,nk,:],0,self.invsmear)

        return self.efermi

    def setFermiFixed(self,broad=1e-5):
        """
        Set Fermi level using fixed occupations method
        Useful for semi-conductors
        """
        eigenvalues = self.eigenvalues_ibz

        #top of valence
        top = np.max(eigenvalues[:,:,self.nbandsv-1])
        #bottom of conduction
        bot = np.max(eigenvalues[:,:,self.nbandsv])
        efermi = (top+bot)/2.
        self.setFermi(efermi,broad)

    def expandEigenvalues(self):
        """
        Expand eigenvalues to the full brillouin zone
        """
        
        self.expand_kpoints()
        self.eigenvalues = self.eigenvalues_ibz[:,self.kpoints_indexes]
        self.nkpoints = len(self.eigenvalues)

    def expand_kpoints(self,verbose=1,atol=1.e-4):
        """ 
        Wrapper for expand_kpoints in kpoints module

        Take a list of qpoints and symmetry operations and return the full brillouin zone
        with the corresponding index in the irreducible brillouin zone
        """

        #check if the kpoints were already exapnded
        if self.expanded == True: return self.kpoints_full, self.kpoints_indexes, self.symmetry_indexes

        weights, kpoints_indexes, symmetry_indexes, kpoints_full = expand_kpoints(self.car_kpoints,self.sym_car,self.rlat,atol=atol)

        #set the variables
        self.expanded = True
        self.kpoints_full     = kpoints_full
        self.kpoints_indexes  = kpoints_indexes
        self.symmetry_indexes = symmetry_indexes
        self.nkpoints         = len(kpoints_full)
        self.weights_ibz      = weights
        self.weights          = np.full((self.nkpoints), 1.0/self.nkpoints,dtype=np.float32)

        if verbose: print("%d kpoints expanded to %d"%(len(self.car_kpoints),len(kpoints_full)))

        return self.kpoints_full, self.kpoints_indexes, self.symmetry_indexes

    def energy_gaps(self,eigenvalues=None,GWshift=0.,nv=-1,verbose=1):
        """
        Calculate the energy of the gap and apply custom rigid shift

        The eigenvalues array [nk,nb] can be given as an argument (useful for QP energies or spin-polarized case),
        otherwise the eigenvalues are taken from the object (spin up channel).
        """
        if eigenvalues is None: eiv = self.eigenvalues_ibz[0]
        else:                   eiv = eigenvalues
        if nv<1: nv = self.nbandsv # call from YamboDipolesDB may have reduced set of bands

        # First apply shift if there is one
        eiv[:,nv:]+=GWshift

        # Then compute gaps
        Egap = np.min(eiv[:,nv]) - np.max(eiv[:,nv-1])
        Edir = np.min(eiv[:,nv]  -        eiv[:,nv-1])

        if verbose:
            print('DFT Energy gap: %s eV'%Egap)
            print('DFT Direct gap: %s eV'%Edir)
            print('GW shift:       %s eV'%GWshift)

        return eiv

    ##
    ## Simple band plot
    ##
    def plot_bs_ax(self,ax,path,bandmin=None,bandmax=None,add_indexes=False,**kwargs):
        """
        Plot this bandstructure on Matpltolib ax
        """
        bands_kpoints, bands_indexes, path_car = get_path(self.car_kpoints,self.rlat,path)
        bands_highsym_qpts = path_car.kpoints
        self.get_fermi()

        #calculate distances
        bands_distances = [0]
        distance = 0
        for nk in range(1,len(bands_kpoints)):
            distance += np.linalg.norm(bands_kpoints[nk]-bands_kpoints[nk-1])
            bands_distances.append(distance)

        #plot high-symmetry qpoints
        distance = 0
        bands_highsym_qpts_distances = [0]
        for nk in range(1,len(bands_highsym_qpts)):
            ax.axvline(distance,color='k')
            distance += np.linalg.norm(bands_highsym_qpts[nk]-bands_highsym_qpts[nk-1])
            bands_highsym_qpts_distances.append(distance)
        ax.axvline(distance,color='k')

        #plot bands
        if self.spin == 2:
           color = kwargs.pop('c','red')
           ax.plot(bands_distances,self.eigenvalues[0,bands_indexes,bandmin:bandmax],c=color,**kwargs)
           color = kwargs.pop('c','blue')
           ax.plot(bands_distances,self.eigenvalues[1,bands_indexes,bandmin:bandmax],c=color,**kwargs)
        else:
            color = kwargs.pop('c','red')
            ax.plot(bands_distances,self.eigenvalues[0,bands_indexes,bandmin:bandmax],c=color,**kwargs)
        
        ax.set_xlim(0,max(bands_distances))

        if add_indexes:
            ax.set_xticks(bands_distances)
            ax.set_xticklabels(np.array(bands_indexes)+1)
            for d in bands_distances:
                ax.axvline(d,color='k',alpha=0.5)

        return ax

    ##
    ## DOS and JDOS
    ##
    def get_transitions(self,eigenvalues=None,nvalence=None,nconduction=None):
        """
        Calculate transition energies
        """
        if eigenvalues is None: eigenvalues = self.eigenvalues_ibz[0] #assume non spin polarized if not given
        if nvalence is None:    nvalence    = self.nbandsv
        if nconduction is None: nconduction = self.nbandsc

        eigs_valence    = eigenvalues[:, :nvalence] 
        eigs_conduction = eigenvalues[:, nvalence:nvalence + nconduction]
        transitions = eigs_conduction[:, None, :] - eigs_valence[:, :, None]
        transitions = transitions.reshape(self.nkpoints_ibz, nvalence * nconduction)

        return transitions

    def getDOS(self,eigenvalues=None,broad=0.1,emin=-10,emax=10,estep=0.01):
        """
        Retrieve the Density of States (DOS)
        (Should work for metals as well but untested)
        """
        if eigenvalues is None: eigenvalues = self.eigenvalues_ibz[0] # assume non spin polarized if not given
        if not self.EXPAND: self.expandEigenvalues()                  # we need this to get the weights
        weights = self.weights_ibz[:self.nkpoints_ibz]
        
        w, dos = get_spectra(eigenvalues,weights=weights,broadening=broad,emin=emin,emax=emax,estep=estep)

        return w, dos

    def getJDOS(self,eigenvalues=None,nconduction=None,broad=0.1,emin=0,emax=10,estep=0.01):
        """
        Calculate the joint density of states
        """
        transitions = self.get_transitions(eigenvalues=eigenvalues,nconduction=nconduction)
        if not self.EXPAND: self.expandEigenvalues() # we need this to get the weights
        weights = self.weights_ibz[:self.nkpoints_ibz]

        w, jdos = get_spectra(transitions,weights=weights,broadening=broad,emin=emin,emax=emax,estep=estep)

        return w, jdos
    
    def setLifetimes(self,broad=0.1):
        """
        Add electronic lifetimes using the DOS
        """
        self.lifetimes_ibz = np.ones(self.eigenvalues_ibz.shape,dtype=np.float32)*broad
        if self.Expand: self.lifetimes = np.ones(self.eigenvalues.shape,dtype=np.float32)*broad

    def setLifetimesDOS(self,broad=0.1,debug=False):
        """
        Approximate the electronic lifetimes using the DOS
        """
        eigenvalues = self.eigenvalues_ibz
        weights = self.weights_ibz
        nkpoints = self.nkpoints_ibz

        #get dos
        emin = np.min(eigenvalues)-broad
        emax = np.max(eigenvalues)+broad
        energies, dos = self.getDOS(emin=emin, emax=emax, step=0.1, broad=broad)

        #normalize dos to broad
        dos = dos/np.max(dos)*broad

        #create a interpolation function to get the lifetimes for all the values
        from scipy.interpolate import interp1d
        f = interp1d(energies, dos, kind='cubic')

        if debug:
            """
            plot the calculated values for the DOS and the interpolated values
            """
            import matplotlib.pyplot as plt
            d=0.001
            x = np.arange(emin-d,emax+d,d)
            plt.plot(energies,dos,'o')
            plt.plot(x,f(x))
            plt.show()
            exit()

        #add imaginary part to the energies proportional to the DOS
        self.lifetimes_ibz = np.array([ [f(eig) for eig in eigk] for eigk in self.eigenvalues_ibz],dtype=np.float32)
        if self.Expand: self.lifetimes = np.array([ [f(eig) for eig in eigk] for eigk in self.eigenvalues],dtype=np.float32)
    
    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app(f"spin polarizations: {self.spin}")
        app(f"spinor components:  {self.spinor_components}")
        app(f"nelectrons: {self.nelectrons}")
        app(f"nbands:     {self.nbands}")
        app(f"nkpoints (IBZ):   {self.nkpoints_ibz}")
        return "\n".join(lines)



#
# This example show how to calculate Joint Density of States using YamboPy
#
from qepy import *
from yambopy import *
import matplotlib.pyplot as plt

save_folder='./SAVE' 
yel = YamboElectronsDB.from_db_file(folder=save_folder)
energies,jdos=yel.getJDOS(emax=30.0)

fig = plt.figure(figsize=(6,4))
plt.xlabel('$Energy[eV]$')
plt.ylabel('JDOS')
plt.plot(energies,jdos)
plt.show()



"""
Tutorial for YamboLatticeDB.

We show a specific functionality, namely kpoint reading and expansion.

EDIT the path below to point to the yambo SAVE folder.
"""
save_path='BSE_saves/YAMBO_saves'
from yambopy import YamboLatticeDB
from yambopy.plot.plotting import BZ_Wigner_Seitz
import numpy as np
import matplotlib.pyplot as plt

if __name__ == "__main__":
    """ 
    Main part of the script
    """
    Cartesian_Plot = True
    Crystal_Plot   = False
    Symmetry_Plot  = False
    i_k = 8 #Custom kpoint index in the IBZ. To be expanded in the full BZ.

    #                    #
    # Start Yambopy part #
    #                    #

    # Create "lattice" object by reading the ns.db1 database inside the yambo SAVE
    ylat = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1')
    print(ylat)

    # Read it again without expansion of the kpts from IBZ to BZ
    ylat_noexp = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1',Expand=False)

    # Apply symmetries: we select one point in the IBZ and manually expand it by applying
    #                   all symmetries in Cartesian coordinates
    kpt = ylat_noexp.car_kpoints[i_k]
    klist = np.array([ np.matmul(symmetry, kpt) for symmetry in ylat_noexp.sym_car  ])

    #                                                                #
    # End Yambopy part. After this is just plotting with matplotlib. #
    #                                                                #

    if Cartesian_Plot:
        # Do a 2D scatterplot of the kpoints in Cartesian coordinates
        fig = plt.figure(figsize=(9,9))
        ax = plt.gca()
        ax.set_title('Cartesian coordinates')

        ## Add BZ borders
        ax.add_patch(BZ_Wigner_Seitz(ylat,color='black',linewidth=1.))

        ## Plot with "nice" layout
        ax.set_aspect('equal')
        ax.scatter(ylat.car_kpoints[:,0],ylat.car_kpoints[:,1],marker='H',s=200,color='teal',\
                   linewidth=0.5,edgecolors='black',label='expanded')
        ax.scatter(ylat_noexp.car_kpoints[:,0],ylat_noexp.car_kpoints[:,1],marker='h',s=100,color='orange',\
                   linewidth=0.5,edgecolors='black',label='unexpanded')

        ## Explicitly show kpt indices
        for i_k,kpt in enumerate(ylat.car_kpoints): 
            kx,ky = kpt[0],kpt[1]
            ax.annotate(i_k, (kx,ky), color='teal', xytext=(kx+0.003,ky+0.005))
        for i_k,kpt in enumerate(ylat_noexp.car_kpoints): 
            kx,ky = kpt[0],kpt[1]
            ax.annotate(i_k, (kx,ky), color='orange', xytext=(kx-0.005,ky+0.005))

        plt.legend()
        plt.show()

    if Crystal_Plot:
        # Do a 2D scatterplot of the kpoints in crystal coordinates
        fig = plt.figure(figsize=(9,9))
        ax = plt.gca()
        ax.set_title('Crystal coordinates')

        ax.set_aspect('equal')
        ax.scatter(ylat.red_kpoints[:,0],ylat.red_kpoints[:,1],marker='H',s=200,color='teal',\
                   linewidth=0.5,edgecolors='black',label='expanded')
        ax.scatter(ylat_noexp.red_kpoints[:,0],ylat_noexp.red_kpoints[:,1],marker='h',s=100,color='orange',\
                   linewidth=0.5,edgecolors='black',label='unexpanded')

        plt.legend()
        plt.show()

    if Symmetry_Plot:
        # Plot star of kpt
        fig = plt.figure(figsize=(9,9))
        ax = plt.gca()
        ax.set_title('Kpoint expansion')
        ax.add_patch(BZ_Wigner_Seitz(ylat,color='black',linewidth=1.))
        ax.set_aspect('equal')
        ax.scatter(ylat.car_kpoints[:,0],ylat.car_kpoints[:,1],marker='H',s=200,color='teal',\
                   linewidth=0.5,edgecolors='black',label='expanded')

        ax.scatter(klist[:,0],klist[:,1],marker='h',s=100,color='orange',\
                   linewidth=0.5,edgecolors='black',label='Custom kpoint')

        plt.legend()
        plt.show()



# Copyright (c) 2023-2025, Claudio Attaccalite,Mike Nico Pionteck
#                          Myrta Gruening
# All rights reserved.
#
# This file is part of the yambopy project
# Calculate linear response from real-time calculations (yambo_nl)
# Modified to allow generalisation
#
import numpy as np
from yambopy.units import ha2ev,fs2aut, Junit, EFunit,SVCMm12VMm1,AU2VMm1
from yambopy.nl.external_efield import Divide_by_the_Field
from tqdm import tqdm
from scipy.ndimage import uniform_filter1d
import sys
import os
import itertools
from abc import ABC,abstractmethod
from yambopy.nl.nl_analysis import Xn_from_signal
#
#
#
# Derived class for frequency-mixed signal
#    
class Xn_from_freqmix(Xn_from_signal):
    #
    def set_defaults(self):
        self.l_out_current = False # there is no implementation for current yet though it can be easily introduced
        if self.solver == '':
            self.solver = 'svd'
        if self.samp_mod == '':
            self.samp_mod = 'log'
        EFIELDS = ["SIN","SOFTSIN"]
        if self.efield["name"] not in EFIELDS:
            raise ValueError(f"Invalid electric field for frequency mixing analysis. Expected one of: {EFIELDS}")
        EFIELDS = ["","SIN","SOFTSIN"]
        if self.pumps[0]["name"] not in EFIELDS:
            raise ValueError(f"Invalid pump field for frequency mixing analysis. Expected one of: {EFIELDS}")
        self.pump_freq = 0.0
        if self.pumps[0]["name"] != "none":
            self.pump_freq = self.pumps[0]["freq_range"][0]
        if len(self.pumps)>1 and self.pumps[1]["name"] != "none":
            raise ValueError("Only mixing of two fields implemented.")
        if isinstance (self.X_order,int): # for frequency mixing it expects 2 orders, if one is given the same is given
            self.X_order = [self.X_order,self.X_order]
            if self.pump_freq == 0.0:
                self.X_order[1] = 0
        self.out_dim = (2*self.X_order[0] + 1)*(2*self.X_order[1] + 1)
    #
    def update_time_range(self):
        T_range = self.T_urange
        if T_range[0] <= 0.0:
            T_range[0] = self.T_deph
        if T_range[1] <= 0.0:
            T_range[1]=self.time[-1]
        return T_range
    #    
    def set_sampling(self,ifrq):
        samp_order = self.out_dim
        if self.nsamp == -1 or self.nsamp < samp_order:
            self.nsamp = int(samp_order**2)
        #
        T_range = self.update_time_range()
        return T_range

    def define_matrix(self,T_i,ifrq):
        NX,MX = self.X_order[:]
        W1 = self.freqs[ifrq]
        W2 = self.pump_freq
        M = np.zeros((self.nsamp, self.out_dim), dtype=np.cdouble)
        for i_t in range(self.nsamp):
            for i_c,(i_n,i_m) in enumerate(itertools.product(range(-NX, NX+1),range(-MX, MX+1))):
                M[i_t, i_c] = np.exp(-1j * (i_n*W1+i_m*W2) * T_i[i_t],dtype=np.cdouble)
        return M

    def output_analysis(self,out,to_file=True):
        #
        NX,MX = self.X_order[:]
        T, _ = self.update_time_range()
        run_info = self.append_runinfo(T)
        #
        for i_f in range(self.n_runs):
            for i_c,(i_n,i_m) in enumerate(itertools.product(range(-NX, NX+1),range(-MX, MX+1))):
                field_fac = 1.0
                if i_n !=0: field_fac *= Divide_by_the_Field(self.efields[i_f],abs(i_n)) #check this
                if self.pump_freq !=0:
                    if i_m !=0: field_fac *= Divide_by_the_Field(self.pumps[0],abs(i_m)) #check this! what is nldb.Efield2?
                out[i_c,i_f,:] *= field_fac
                out[i_c,i_f,:] *= self.get_Unit_of_Measure(abs(i_n)+abs(i_m))
        if (to_file):
            for i_c,(i_n,i_m) in enumerate(itertools.product(range(-NX, NX+1),range(-MX, MX+1))):
                output_file = f'o{self.prefix}.YamboPy-X_probe_order_{i_n}_{i_m}'
                iorder = (i_n+i_m,i_n,i_m)
                header = "E[eV] " + " ".join([f"X{iorder}/Im_{d} X{iorder}/Re_{d}" for d in ('x','y','z')])
                values = np.column_stack((self.freqs * ha2ev, out[i_c, :, 0].imag, out[i_c, :, 0].real,
                        out[i_c, :, 1].imag, out[i_c, :, 1].real,
                        out[i_c, :, 2].imag, out[i_c, :, 2].real))
                np.savetxt(output_file, values, header=header, delimiter=' ', footer=f"Frequency mixing analysis with pump frequency {self.pump_freq*ha2ev} eV")
                outf = open(output_file, "a")
                outf.writelines(run_info)
                outf.close()
        else:
            print(run_info)
            return (self.freqs, out)

    def reconstruct_signal(self,out,to_file=True):
        NX,MX = self.X_order[:]
        Seff = np.zeros((self.n_runs, 3, len(self.time)), dtype=np.cdouble)
        for i_f in tqdm(range(self.n_runs)):
            for i_d in range(3):
                for i_c,(i_n,i_m) in enumerate(itertools.product(range(-NX, NX+1),range(-MX, MX+1))):
                    freq_term = np.exp(-1j * (i_n * self.freqs[i_f] + i_m*self.pump_freq) * self.time)
                    Seff[i_f, i_d, :] += out[i_c, i_f, i_d] * freq_term
        if (to_file):
            for i_f in tqdm(range(self.n_runs)):
                values = np.column_stack((self.time / fs2aut, Seff[i_f, 0, :].real, Seff[i_f, 1, :].real, Seff[i_f, 2, :].real))
                output_file = f'o{self.prefix}.YamboPy-pol_reconstructed_F{i_f + 1}'
                header="[fs] Px Py Pz"
                np.savetxt(output_file, values, header=header, delimiter=' ', footer="Reconstructed signal")
        else:
            return values
    #
    # unused function so far
    # check for spikes in the solution and recalculate the solution
    # The function can be called after perform_analysis
    # in a loop over cartesian directions
    # Note that it can be taken into nl_analysis to be shared with other classes
    #    
    def spike_correction(self,mat,samp,out):
        window_size, threshold_local = (5,2.0)
        out_corr = out
        self.solver = "lstq_opt"
        signal_im= abs(out[self.X_order[0]+1,self.X_order[1]+1,:].imag)
        signal_re= abs(out[self.X_order[0]+1,self.X_order[1]+1,:].real)
        smooth_signal_im = uniform_filter1d(signal_im, size=window_size)
        smooth_signal_re = uniform_filter1d(signal_re, size=window_size)
        spike_indices_local_im = np.where(np.abs(signal_im - smooth_signal_im)/smooth_signal_im > threshold_local)[0]
        spike_indices_local_re = np.where(np.abs(signal_re - smooth_signal_re)/smooth_signal_re > threshold_local)[0]
        spike_indices_local = np.unique(np.concatenate((spike_indices_local_im, spike_indices_local_re)))
        for i_f in tqdm(spike_indices_local):
            if(i_f==0 and i_f in spike_indices_local):
                x0=out[:,i_f+1]
            elif(i_f==n_frequencies-1 and i_f in spike_indices_local):
                x0=out[:,i_f-1]
            else:
                x0=(out[:,i_f+1]+out[:,i_f-1])/2.0
                out_corr[:,i_f] = self.solve_lin_system(mat,samp,init=x0)
        return out_corr, spike_indices_local



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
import xml.etree.ElementTree as ET
from qepy.auxiliary import *
from .lattice import *
from yambopy.plot.plotting import add_fig_kwargs 
from re import findall
from numpy import zeros

__all__ = ['PwXML']

HatoeV = 27.2107

class PwXML():
    # This class reads up to version 6.7
    """ Class to read data from a Quantum espresso XML file
    """
    _eig_xml   = 'eigenval.xml'
    _eig1_xml  = 'eigenval1.xml'
    _eig2_xml  = 'eigenval2.xml'

    
    def __init__(self,prefix,path='.',verbose=0):
        """ Initialize the structure with the path where the datafile.xml is
        """
        self.prefix = prefix
        self.path   = path

        datafiles = {'data-file.xml':        self.read_datafile,
                     'data-file-schema.xml': self.read_datafile_schema}

        done_reading = False

        #check if the name is data-file.xml or data-file-schema.xml or whatever....
        for filename,read in list(datafiles.items()):
            path_filename = "%s/%s.save/%s"%(path, prefix, filename)
            if os.path.isfile(path_filename):
                if verbose: print("reading %s"%filename)
                done_reading = read(path_filename)
                break

        #trap errors
        if not done_reading:
            possible_files = " or ".join(list(datafiles.keys()))
            raise ValueError('Failed to read %s in %s/%s.save'%(possible_files,path,prefix))

    def read_datafile(self,filename):
        """
        Read some data from the xml file in the old format of quantum espresso
        """
        self.datafile_xml = ET.parse( filename ).getroot()


        #get magnetization state

        self.lsda = False
        if 'T' in self.datafile_xml.findall("SPIN/LSDA")[0].text:
            self.lsda = True

        #get acell
        self.celldm = [ float(x) for x in self.datafile_xml.findall("CELL/CELL_DIMENSIONS")[0].text.strip().split('\n') ]

        #get cell
        self.cell = []
        for i in range(1,4):
            cell_lat = self.datafile_xml.findall("CELL/DIRECT_LATTICE_VECTORS/a%d"%i)[0].text
            self.cell.append([float(x) for x in cell_lat.strip().split()])

        #get reciprocal cell
        self.rcell = []
        for i in range(1,4):
            rcell_lat = self.datafile_xml.findall("CELL/RECIPROCAL_LATTICE_VECTORS/b%d"%i)[0].text
            self.rcell.append([float(x) for x in rcell_lat.strip().split()])

        #get atoms
        self.natoms = int(self.datafile_xml.findall("IONS/NUMBER_OF_ATOMS")[0].text)
        self.atoms = []
        for i in range(1,self.natoms+1):
            atom_xml = self.datafile_xml.findall("IONS/ATOM.%d"%i)[0]
            #read positions
            pos_string = atom_xml.get('tau').strip().split()
            pos = [float(x) for x in pos_string]
            #read type  
            atype = atom_xml.get('SPECIES').strip()
            #store
            self.atoms.append([atype,pos])

        #get atomic species
        self.natypes = int(self.datafile_xml.findall("IONS/NUMBER_OF_SPECIES")[0].text) 
        self.atypes = {}
        for i in range(1,self.natypes+1):
            atype_xml = self.datafile_xml.findall("IONS/SPECIE.%d"%i)[0]
            #read string
            atype_string = atype_xml.findall('ATOM_TYPE')[0].text.strip()
            #read mass
            atype_mass =  float(atype_xml.findall('MASS')[0].text.strip())
            #read pseudo
            atype_pseudo =  atype_xml.findall('PSEUDO')[0].text.strip()
            self.atypes[atype_string]=[atype_mass,atype_pseudo]

        #get nkpoints

        #get nkpoints
        self.nkpoints = int(self.datafile_xml.findall("BRILLOUIN_ZONE/NUMBER_OF_K-POINTS")[0].text.strip())
        # Read the number of BANDS
        self.nbands   = int(self.datafile_xml.find("BAND_STRUCTURE_INFO/NUMBER_OF_BANDS").text)

        #get k-points
        self.kpoints = [] 
        for i in range(self.nkpoints):
          k_aux = self.datafile_xml.findall('BRILLOUIN_ZONE/K-POINT.%d'%(i+1))[0].get('XYZ')
          self.kpoints.append([float(x) for x in k_aux.strip().split()])
        
        #get fermi
        self.fermi = float(self.datafile_xml.find("BAND_STRUCTURE_INFO/FERMI_ENERGY").text)*HatoeV
 
        #get eigenvalues

        if not self.lsda:

           eigen = []
           for ik in range(self.nkpoints):
               for EIGENVALUES in ET.parse( "%s/%s.save/K%05d/%s" % (self.path,self.prefix,(ik + 1),self._eig_xml) ).getroot().findall("EIGENVALUES"):
                   eigen.append(list(map(float,EIGENVALUES.text.split())*HatoeV))
           self.eigen  = eigen - self.fermi
           self.eigen1 = eigen - self.fermi

        #get eigenvalues of spin up & down

        if self.lsda:
           eigen1, eigen2 = [], []
           for ik in range(self.nkpoints):
               for EIGENVALUES1 in ET.parse( "%s/%s.save/K%05d/%s" % (self.path,self.prefix,(ik + 1),self._eig1_xml) ).getroot().findall("EIGENVALUES"):
                    eigen1.append(list(map(float, EIGENVALUES1.text.split())*HatoeV))
               for EIGENVALUES2 in ET.parse( "%s/%s.save/K%05d/%s" % (self.path,self.prefix,(ik + 1),self._eig2_xml) ).getroot().findall("EIGENVALUES"):
                    eigen2.append(list(map(float, EIGENVALUES2.text.split())*HatoeV))

           self.eigen   = eigen1 - self.fermi
           self.eigen1  = eigen1 - self.fermi
           self.eigen2  = eigen2 - self.fermi

        #get occupations of spin up & down
        if self.lsda:
           occ1, occ2 = [], []
           for ik in range(self.nkpoints):
               for OCCUPATIONS1 in ET.parse( "%s/%s.save/K%05d/%s" % (self.path,self.prefix,(ik + 1),self._eig1_xml) ).getroot().findall("OCCUPATIONS"):
                    occ1.append(list(map(float, OCCUPATIONS1.text.split())))
               for OCCUPATIONS2 in ET.parse( "%s/%s.save/K%05d/%s" % (self.path,self.prefix,(ik + 1),self._eig2_xml) ).getroot().findall("OCCUPATIONS"):
                    occ2.append(list(map(float, OCCUPATIONS2.text.split())))
        
           self.occupation1 = occ1
           self.occupation2 = occ2

        #get Bravais Lattice
        self.bravais_lattice = str(self.datafile_xml.find("CELL/BRAVAIS_LATTICE").text)
        if all(s in self.bravais_lattice for s in ["cubic","P"]):               self.ibrav = 1
        if all(s in self.bravais_lattice for s in ["cubic","F"]):               self.ibrav = 2
        if all(s in self.bravais_lattice for s in ["cubic","I"]):               self.ibrav = 3
        if all(s in self.bravais_lattice for s in ["Hexagonal","Trigonal"]):    self.ibrav = 4
        if all(s in self.bravais_lattice for s in ["Trigonal","R"]):            self.ibrav = 5
        if all(s in self.bravais_lattice for s in ["Tetragonal","P"]):          self.ibrav = 6
        if all(s in self.bravais_lattice for s in ["Tetragonal","I"]):          self.ibrav = 7
        if all(s in self.bravais_lattice for s in ["Orthorhombic","P"]):        self.ibrav = 8
        if all(s in self.bravais_lattice for s in ["Orthorhombic","base-centered"]): self.ibrav = 9
        if all(s in self.bravais_lattice for s in ["Orthorhombic","face-centered"]): self.ibrav = 10
        if all(s in self.bravais_lattice for s in ["Orthorhombic","body-centered"]): self.ibrav = 11
        if all(s in self.bravais_lattice for s in ["Monoclinic","P"]):          self.ibrav = 12
        if all(s in self.bravais_lattice for s in ["Monoclinic","base-centered"]):   self.ibrav = 13
        if all(s in self.bravais_lattice for s in ["Triclinic"]):               self.ibrav = 14


        return True

    def read_datafile_schema(self,filename):
        """
        Read the data from the xml file in the new format of quantum espresso
        """
        self.datafile_xml = ET.parse( filename ).getroot()

        # occupation type

        self.occ_type = self.datafile_xml.findall("input/bands/occupations")[0].text

        #get magnetization state
        # TO BE DONE!!!
        self.lsda = False
        if 'true' in self.datafile_xml.findall("input/spin/lsda")[0].text:
            self.lsda = True
            #raise ValueError('Spin states not yet implemented for data-file-schema.xml') 

        #get cell
        self.cell = []
        for i in range(1,4):
            cell_lat = self.datafile_xml.findall("output/atomic_structure/cell/a%d"%i)[0].text
            self.cell.append([float(x) for x in cell_lat.strip().split()])

        #calculate acell
        self.acell = [ np.linalg.norm(a) for a in self.cell ]

        #get reciprocal cell
        self.rcell = []
        for i in range(1,4):
            rcell_lat = self.datafile_xml.findall("output/basis_set/reciprocal_lattice/b%d"%i)[0].text
            self.rcell.append([float(x) for x in rcell_lat.strip().split()])

        #get atoms
        self.natoms = int(self.datafile_xml.findall("output/atomic_structure")[0].get('nat'))
        self.atoms = []
        atoms = self.datafile_xml.findall("output/atomic_structure/atomic_positions/atom")
        for i in range(self.natoms):
            atype = atoms[i].get('name')
            pos = [float(x) for x in atoms[i].text.strip().split()]
            self.atoms.append([atype,pos])

        #get atomic species
        self.natypes = int(self.datafile_xml.findall("output/atomic_species")[0].get('ntyp')) 
        atypes = self.datafile_xml.findall("output/atomic_species/species")
        self.atypes = {}
        for i in range(self.natypes):
            #read string
            atype_string = atypes[i].get('name').strip()
            #read mass
            atype_mass = atypes[i].findall('mass')[0].text.strip()
            #read pseudo
            atype_pseudo = atypes[i].findall('pseudo_file')[0].text.strip()
            self.atypes[atype_string]=[atype_mass,atype_pseudo]

        #get nkpoints
        self.nkpoints = int(self.datafile_xml.findall("output/band_structure/nks")[0].text.strip())
        # Read the number of BANDS
        if self.lsda:
           self.nbands_up = int(self.datafile_xml.findall("output/band_structure/nbnd_up")[0].text.strip())
           self.nbands_dw = int(self.datafile_xml.findall("output/band_structure/nbnd_dw")[0].text.strip())
           self.nbands = self.nbands_up + self.nbands_dw
        else:
           self.nbands = int(self.datafile_xml.findall("output/band_structure/nbnd")[0].text.strip())

        #get ks states
        kstates = self.datafile_xml.findall('output/band_structure/ks_energies')

        #get k-points
        self.kpoints = [] 
        for i in range(self.nkpoints):
            kpoint = [float(x) for x in kstates[i].findall('k_point')[0].text.strip().split()]
            self.kpoints.append( kpoint )

        #get fermi (it depends on the occupations and spin pol)
        if self.occ_type == 'fixed':
           self.fermi = float(self.datafile_xml.find("output/band_structure/highestOccupiedLevel").text)*HatoeV
        else:
            try: 
                self.fermi = float(self.datafile_xml.find("output/band_structure/fermi_energy").text)*HatoeV
            except AttributeError: 
                fermis = self.datafile_xml.find("output/band_structure/two_fermi_energies").text.split()
                self.fermis = [float(fermis[0])*HatoeV,float(fermis[1])*HatoeV]
                self.fermi = self.fermis[1] # set to spin minority energy

        #get eigenvalues
        self.eigen1 = []
        for k in range(self.nkpoints):
            eigen = [float(x) for x in kstates[k].findall('eigenvalues')[0].text.strip().split()]
            self.eigen1.append( eigen )
        self.eigen1 = np.array(self.eigen1)*HatoeV - self.fermi
 
        #get Bravais lattice
        self.ibrav = self.datafile_xml.findall("output/atomic_structure")[0].get('bravais_index')

        return True

    def get_scaled_atoms(self):
        """Build and return atoms dictionary"""
        atoms = []
        for atype,pos in self.atoms:
            red_pos = car_red([pos],self.cell)[0].tolist()
            atoms.append([atype,red_pos])
        return atoms
        
    def get_atypes_dict(self):
        return self.atypes

    def get_lattice_dict(self):
        return dict(ibrav=self.ibrav,cell_parameters=self.cell)

    def get_structure_dict(self):
        """
        Get a structure dict that can be used to create a new pw input file
        """
        return dict(atoms=self.get_scaled_atoms(),
             atypes=self.get_atypes_dict(),
             lattice=self.get_lattice_dict())

    def get_cartesian_positions(self):
        """ get the atomic positions in cartesian coordinates """
        positions = []
        for atype,pos in self.atoms:
            positions.append(pos)
        return np.array(positions)

    def get_scaled_positions(self):
        """ get the atomic positions in reduced coordinates """
        cartesian_positions = self.get_cartesian_positions()
        return car_red(cartesian_positions,self.cell)

    def __str__(self):
        lines = []; app = lines.append
        app("cell:")
        for c in self.cell:
            app(("%12.8lf "*3)%tuple(c))
        app("atoms:")
        for atype,pos in self.atoms:
            app("%3s"%atype + ("%12.8lf "*3)%tuple(pos))
        app("nkpoints: %d"%self.nkpoints)
        app("nbands:   %d"%self.nbands)
        return "\n".join(lines)

    def plot_eigen_ax(self,ax,path_kpoints=[],xlim=(),ylim=(),color='r',**kwargs):
        #
        # Careful with variable path. I am substituting vy path_kpoints
        # To be done in all the code (and in the tutorials)
        #
        # argurments:
        # ls: linestyle
        if path_kpoints:
            if isinstance(path_kpoints,Path):
                path_kpoints = path_kpoints.get_indexes()
                path_ticks, path_labels = list(zip(*path_kpoints))
            ax.set_xticks( path_ticks )
            ax.set_xticklabels( path_labels )
        ax.set_ylabel('E (eV)')

        ls = kwargs.pop('ls','solid')
        lw = kwargs.pop('lw',1)
        y_offset = kwargs.pop('y_offset',0.0)
        #get kpoint_dists 
        kpoints_dists = calculate_distances(self.kpoints)
        ticks, labels = list(zip(*path_kpoints))
        ax.set_xticks([kpoints_dists[t] for t in ticks])
        ax.set_xticklabels(labels)
        ax.set_xlim(kpoints_dists[0],kpoints_dists[-1])

        #plot vertical lines
        for t in ticks:
            ax.axvline(kpoints_dists[t],c='k',lw=2)
        ax.axhline(0,c='k')

        #plot bands
       
        if self.lsda:
           eigen1 = np.array(self.eigen1)

           for ib in range(self.nbands_up):
               ax.plot(kpoints_dists,eigen1[:,ib]                + y_offset, '%s-'%color, lw=lw, zorder=1,label='spin-up') # spin-up 
               ax.plot(kpoints_dists,eigen1[:,ib+self.nbands_up] + y_offset, 'b-', lw=lw, zorder=1,label='spin-down') # spin-down

           import matplotlib.pyplot as plt
           handles, labels = plt.gca().get_legend_handles_labels()
           by_label = dict(zip(labels, handles))
           plt.legend(by_label.values(), by_label.keys())

        # Case: Non spin polarization
        else:
           eigen1 = np.array(self.eigen1)

           for ib in range(self.nbands):
               ax.plot(kpoints_dists,eigen1[:,ib] + y_offset, color=color,linestyle=ls , lw=lw, zorder =1)

        #plot options
        if xlim: ax.set_xlim(xlim)
        if ylim: ax.set_ylim(ylim)

     
    #def plot_eigen_spin_ax(self,ax,path_kpoints=[],xlim=(),ylim=(),spin_proj=None):
    def plot_eigen_spin_ax(self,ax,path_kpoints=[],xlim=(),ylim=(),spin_proj=False,spin_folder='.'):
        #
        # Careful with variable path. I am substituting vy path_kpoints
        # To be done in all the code (and in the tutorials)
        # This is a test function for spin-polarized bands
        #
        #
        
        import matplotlib.pyplot as plt
        self.spin_proj = np.array(spin_proj) if spin_proj is not None else None 

        if spin_proj == True:
           self.spin_projection(spin_dir=3,folder=spin_folder,prefix='bands')

        if path_kpoints:
            if isinstance(path_kpoints,Path):
                path_kpoints = path_kpoints.get_indexes()
                path_ticks, path_labels = list(zip(*path_kpoints))
            ax.set_xticks( path_ticks )
            ax.set_xticklabels( path_labels )
        ax.set_ylabel('E (eV)')

        # I choose a colormap for spin
        color_map  = plt.get_cmap('seismic')

        #get kpoint_dists 
        kpoints_dists = calculate_distances(self.kpoints)
        ticks, labels = list(zip(*path_kpoints))
        ax.set_xticks([kpoints_dists[t] for t in ticks])
        ax.set_xticklabels(labels)
        ax.set_xlim(kpoints_dists[0],kpoints_dists[-1])

        # NOT WORKING, CHECK IT!
        #plot vertical lines
        #for t in ticks:
        #    ax.axvline(kpoints_dists[t],c='k',lw=2)
        #ax.axhline(0,c='k')

        #plot bands
        eigen1 = np.array(self.eigen1)
        for ib in range(self.nbands):
            x = kpoints_dists
            y = eigen1[:,ib] - self.fermi

            color_spin = self.spin_proj[:,ib] + 0.5 # I renormalize 0 => down; 1 => up
            ax.scatter(x,y,s=100,c=color_spin,cmap=color_map,vmin=0.0,vmax=1.0,edgecolors='none')
       
        #plot spin-polarized bands: TO BE DONE
        #if self.lsda:

        #  eigen2 = np.array(self.eigen2)
        #   for ib in range(self.nbands):
        #       ax.plot(kpoints_dists,eigen2[:,ib]*HatoeV - self.fermi*HatoeV, 'b-', lw=2)

        #plot options
        if xlim: ax.set_xlim(xlim)
        if ylim: ax.set_ylim(ylim)


    '''
    Workaround to include occupations in the plot. AMS
    '''

    def plot_eigen_occ_ax(self,ax,path_kpoints=[],xlim=(),ylim=(),color='r'):

        if path_kpoints:
            if isinstance(path_kpoints,Path):
                path_kpoints = path_kpoints.get_indexes()
            ax.set_xticks( *list(zip(*path_kpoints)) )
        ax.set_ylabel('E (eV)')

        #get kpoint_dists 
        kpoints_dists = calculate_distances(self.kpoints)
        ticks, labels = list(zip(*path_kpoints))
        ax.set_xticks([kpoints_dists[t] for t in ticks])
        ax.set_xticklabels(labels)
        ax.set_xlim(kpoints_dists[0],kpoints_dists[-1])

        #plot vertical lines
        for t in ticks:
            ax.axvline(kpoints_dists[t],c='k',lw=2)
        ax.axhline(0,c='k')
        import matplotlib.pyplot as plt

        #plot bands
        eigen1 = np.array(self.eigen1)
        occ1   = np.array(self.occupation1)
        for ib in range(self.nbands):
            plt.scatter(kpoints_dists,eigen1[:,ib] - self.fermi, s=10*occ1[:,ib],c=color)
       
        #plot spin-polarized bands
        if self.lsda:

           eigen2 = np.array(self.eigen2)
           occ2   = np.array(self.occupation1)
           for ib in range(self.nbands):
               plt.scatter(kpoints_dists,eigen2[:,ib] - self.fermi, s=10*occ2[:,ib],c='b')


        #plot options
        if xlim: ax.set_xlim(xlim)
        if ylim: ax.set_ylim(ylim)

    @add_fig_kwargs
    def plot_eigen(self,path_kpoints=[],xlim=(),ylim=()):
        """ plot the eigenvalues using matplotlib
        """
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        self.plot_eigen_ax(ax,path_kpoints=path_kpoints)
        return fig

    def write_eigen(self,fmt='gnuplot'):
        """ write eigenvalues to a text file
        """
        if fmt=='gnuplot':
            f = open('%s.dat'%self.prefix,'w')
            for ib in range(self.nbands):
                for ik in range(self.nkpoints):
                    f.write("%.1lf %.4lf \n " % (ik,self.eigen1[ik][ib]) )
                f.write("\n")
            f.close()
        else:
            print('fmt %s not implemented'%fmt)

    def spin_projection(self,spin_dir=3,folder='.',prefix='bands'):
        """
        This function reads the spin projection given by bands.x in txt file
        lsigma(i) = .true.
        By default I set the spin direction z ==3

        """
        if spin_dir ==3:  
        #data_eigen  = open('%s/%s.out'  % (folder,prefix),'r').readlines()
           data_spin_3 = open('%s/%s.out.3'% (folder,prefix),'r').readlines()
           
           # check consistency file from bands.x and xml file
           nband = int(findall(r"[-+]?\d*\.\d+|\d+", data_spin_3[0].strip().split()[2]  )[0])
           nk    = int(data_spin_3[0].strip().split()[-2])
           nline = int(nband/10)
           if nband < 10: print("Error, uses only nband => 10 and multiple of 10")
           if self.nbands != nband or self.nkpoints != nk: print("Warning: Dimensions are different!")

           self.spin_3 = zeros([self.nkpoints,self.nbands])

        for ik in range(self.nkpoints):
            for ib in range(nline):
                ib1, ib2, ib3 = int(ib*10), int((ib+1)*10), int(ik*(nband/10+1)+2+ib)
                self.spin_3[ik,ib1:ib2] = list( map(float,data_spin_3[ib3].split()))

    def read_symmetries(self):
        """
        Read symmetry operations from data-file-schema.xml

        Works for ibrav>0

        NB: data-file-schema.xml has nrot and nsym with nsym<nrot.
            They are stored in order with nsym first.

        NB2: Most likely not working with symmorphic symmetries
        """
        #nsym
        nrot = int(self.datafile_xml.findall("output/symmetries/nrot")[0].text.strip())        
        self.nsym = int(self.datafile_xml.findall("output/symmetries/nsym")[0].text.strip())        
        no_t_rev = ( self.datafile_xml.findall("input/symmetry_flags/no_t_rev")[0].text.strip() == "true" )
        self.nsym_with_trev = 2*self.nsym 

        # data
        if not no_t_rev: sym_red = np.zeros((self.nsym_with_trev,3,3))
        else:            sym_red = np.zeros((self.nsym,3,3))
        symmetries = self.datafile_xml.findall("output/symmetries/symmetry/rotation") # All rotations
        symmetries = symmetries[:self.nsym] # Retain only point group symmetries with no trev
        #sym_info   = self.datafile_xml.findall("output/symmetries/symmetry/info")
        #NB: sym_info[:].attrib['class'] contains irrep names if found

        #read (non-symmorphic) symmetries
        for i in range(self.nsym):
            sym = np.array( [float(x) for x in symmetries[i].text.strip().split()] ).reshape(3,3)
            sym_red[i] = sym.T # symmetries are saved as the transposed in the .xml 
        self.sym_red = sym_red.astype(int)

        #convert to c.c.
        self.sym_car = self.sym_red_car()

        #check for time reversal and apply it to sym_car
        if not no_t_rev: self.apply_t_rev()

    def apply_t_rev(self):
        """
        Add T*S rotation matrices
        """
        for n in range(self.nsym,self.nsym_with_trev): self.sym_car[n]=-1*self.sym_car[n-self.nsym]

    def sym_red_car(self):
        """
        Transform symmetry ops. in Cartesian coordinates
        """
        lat = np.array(self.cell)
        sym_car = np.zeros([len(self.sym_red),3,3],dtype=float)
        for n,s in enumerate(self.sym_red):
            sym_car[n] = np.dot( np.linalg.inv(lat), np.dot(s, lat ) ).T
        return sym_car

    def expand_kpoints_xml(self,atol=1e-6,expand_eigen=True,verbose=0):
        """
        Take a list of kpoints and symmetry operations and return the full brillouin zone
        with the corresponding index in the irreducible brillouin zone

        Expand also eigenvalues by default
        """
        alat      = np.array(self.acell)
        kpts_ibz  = np.array(self.kpoints)
        eigen_ibz = np.array(self.eigen1)
        rlat      = np.array(self.rcell)

        self.read_symmetries()

        #check if the kpoints were already exapnded
        kpoints_indices  = []
        kpoints_full     = []
        symmetry_indices = []

        #kpoints in the full brillouin zone organized per index
        kpoints_full_i = {}

        #expand using symmetries
        for nk,k in enumerate(kpts_ibz):
            #if the index in not in the dictionary add a list
            if nk not in kpoints_full_i:
                kpoints_full_i[nk] = []

            for ns,sym in enumerate(self.sym_car):

                new_k = np.dot(sym,k)

                #check if the point is inside the bounds
                k_red = car_red([new_k],rlat)[0]
                k_bz = (k_red+atol)%1

                #if the vector is not in the list of this index add it
                if not vec_in_list(k_bz,kpoints_full_i[nk]):
                    kpoints_full_i[nk].append(k_bz)
                    kpoints_full.append(new_k)
                    kpoints_indices.append(nk)
                    symmetry_indices.append(ns)
                    continue

        #calculate the weights of each of the kpoints in the irreducible brillouin zone
        nkpoints_full = len(kpoints_full)
        weights = np.zeros([nkpoints_full])
        for nk in kpoints_full_i:
            weights[nk] = float(len(kpoints_full_i[nk]))/nkpoints_full

        if verbose: print("%d kpoints expanded to %d"%(self.nkpoints,len(kpoints_full)))

        #set the variables
        self.weights_ibz      = np.array(weights)
        self.kpoints_indices  = np.array(kpoints_indices)
        self.symmetry_indices = np.array(symmetry_indices)
        self.nkbz             = nkpoints_full
        #cartesian coordinates of QE
        self.kpoints_bz       = np.array(kpoints_full)

        if expand_eigen:

            self.eigen_bz = np.zeros((self.nkbz,self.nbands))
            for ik in range(self.nkbz): self.eigen_bz[ik,:] = eigen_ibz[self.kpoints_indices[ik],:]
            if verbose: print("Eigenvalues also expanded.")



import os
from netCDF4 import *
from glob import glob
import argparse

"""
Script to update serial numbers of yambo ndb.* databases in order to import them to new calculations.

Inputs:
 1. --new_serial='path/to/folder/with/new/dbs' [e.g., the new SAVE]
 2. --old_serial='path/to/folder/with/old/dbs' [e.g., an old ndb.em1s]

This script will prompt the user to go through with updating the dbs.
"""

def get_serials(fldr):
    """
    Identify the databases involved in the change and read their serial numbers
    """
    ndbs_tmp = glob('%s/ndb.*'%fldr)
    ns_tmp   = glob('%s/ns.*'%fldr)
    ndbs = []
    for ndb in ndbs_tmp:
        if "fragment" not in ndb: ndbs.append(ndb)
    for ns in ns_tmp:
        if "fragment" not in ns: ndbs.append(ns)

    ndbs_target = []
    ndbs_values = []
    for ndb in ndbs:
        dbs = Dataset(ndb)
        if 'SERIAL_NUMBER' in dbs.variables:
            ndbs_target.append(ndb)
            serial = int(dbs['SERIAL_NUMBER'][0])
            ndbs_values.append(str(serial))
        dbs.close()
    return ndbs_target, ndbs_values

def prompt_user(new,old,check=False):
    """
    Print info for the user and ask if they want to go through with the change
    """
    new_fn,new_sn = get_serials(new)
    old_fn,old_sn = get_serials(old)
    SERIAL_NUMBER = new_sn[0]
    print("Serial numbers found:")
    print("=====================")
    print(">> New databases:")
    if '%s/ndb.gops'%new in new_fn: is_gops_there = True
    else: is_gops_there = False
    are_serials_equal = True
    for sn in new_sn:
        if sn != new_sn[0]: are_serials_equal = False
    if not are_serials_equal:
        if not is_gops_there: raise ValueError('Conflicting serial numbers in newer folder!')
        else: print('Conflicting serial numbers in newere folder. The value of ndb.gops will be used.')
    for i in range(len(new_fn)): 
        print('%s: %s'%(new_fn[i],new_sn[i]))
        if is_gops_there and new_fn[i]=='%s/ndb.gops'%new: SERIAL_NUMBER = new_sn[i]
    print("--------------")
    print(">> Old databases:")
    for i in range(len(old_fn)): print('%s: %s'%(old_fn[i],old_sn[i]))
    print("--------------")
    print("=====================")
    if not check:
        usr_inp = input("Serial number %s will be put in the old databases. Proceed ['y','n']? "%SERIAL_NUMBER)
        if usr_inp != 'y':
            print("Serial numbers not updated.")
            exit(0)
        else:
            update_serial_numbers(SERIAL_NUMBER,old_fn)
            print("Serial numbers updated. Check:")
            prompt_user(new,old,check=True)

def update_serial_numbers(SN,db_names):
    """
    Edit the SERIAL_NUMBER variables in the old dbs
    """
    for db_name in db_names:
        dbs = Dataset(db_name,'r+')
        dbs['SERIAL_NUMBER'][0] = float(SN)
        dbs.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Updated serial numbers in yambo databases')
    parser.add_argument('-new','--new_serial', type=str, default="./SAVE", help='<Optional> Path to folder with the newer databases (Default is ./SAVE)')
    parser.add_argument('-old','--old_serial', type=str,help='<Required> Path to folder with the older databases',required=True)
    args = parser.parse_args()

    new_dir = args.new_serial
    old_dir = args.old_serial

    prompt_user(new_dir,old_dir)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function, division
import os
import re
from math import sqrt
import numpy as np
from qepy.auxiliary import *

eVtocm1 = 8065.54429
cm1toeV = 1.0/eVtocm1
ha2ev  = 27.211396132
ev2ha  = 1.0/ha2ev
Thz2cm1 = 33.35641
cm12Thz = 1.0/33.35641

__all__ = [
"DynmatIn",
]

class DynmatIn(object):
    """
    Generate an manipulate quantum espresso input files for matdyn.x
    """
    def __init__(self):
        self.variable = dict()
        self.qpoints     = [] 

    @classmethod
    def from_prefix(cls,prefix,asr='simple'):
        dm = DynmatIn()
        dm['asr'] = "'%s'"%asr
        dm['fildyn'] = "'%s.dyn1'"%prefix
        dm['filout'] = "'%s.modes'"%prefix
        return dm

    def write(self,filename):
        with open(filename,'w') as f:
            f.write(str(self))

    def __str__(self):
        s = '&input\n'
        s += self.stringify_group('',self.variable) #print variable
        if len(self.qpoints) > 0:
          s+=("%d\n"%len(self.qpoints))
          for q in self.qpoints:
            s+=("%12.8lf %12.8lf %12.8lf")%tuple(q[:3])+"\n"
        return s

    def __setitem__(self,key,value):
        self.variable[key] = value

    def __getitem__(self,key):
        return self.variable[key]

    def stringify_group(self, keyword, group):
        if group != {}:
            string='\n'
            for keyword in group:
                string += "%20s = %s\n" % (keyword, group[keyword])
            string += "/\n"
            return string
        else:
            return ''




# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
Scripts to manipulate Quantum Espresso input files

Also able to read output files in xml format (datafile.xml or datafile-schema.xml)

"""
import os

class qepyenv():
    PW = "pw.x"
    PH = "ph.x"
    DYNMAT = "dynmat.x"
    PSEUDODIR = os.path.join(os.path.dirname(__file__),'data','pseudos')
    CONV_THR = 1e-8

from qepy.lattice import *
from qepy.xml import *
from qepy.bravais import *
from qepy.pw import *
from qepy.pwxml import *
from qepy.projwfc import *
from qepy.projwfcxml import *
from qepy.ph import *
from qepy.dynmat import *
from qepy.matdyn import *
from qepy.unfolding import *
from qepy.supercell import *
from qepy.upf_interface.ppupf import *
from qepy.spin_texture import *



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import unittest
import os
from qepy.lattice import Path
from yambopy.kpoints import get_path
from yambopy.dbs.latticedb import YamboLatticeDB
test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','gw_conv')

class TestYamboLatticeDB(unittest.TestCase):

    def test_yambolatticedb(self):
        """ test latticedb """

        #open latticedb
        filename = os.path.join(test_path,'SAVE/ns.db1')
        ydb = YamboLatticeDB.from_db_file(filename)

        #write json file
        ydb.write_json('lattice.json')        
        string1 = str(ydb)

        #read jsonfile
        ydb.from_json_file('lattice.json')
        string2 = str(ydb)
        
        assert string1 == string2

        p = Path([ [[1.0,1.0,1.0],'G'],
                   [[0.0,0.5,0.5],'X'],
                   [[0.0,0.0,0.0],'G'],
                   [[0.5,0.0,0.0],'L']], [20,20,20])
        bands_kpoints, bands_indexes, path_car = get_path(ydb.car_kpoints,ydb.rlat,None,p)

        print(ydb)

    def tearDown(self): 
        os.remove('lattice.json')

if __name__ == '__main__':
    unittest.main()
